#
# PySNMP MIB module SNMPv2-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///usr/share/snmp/mibs/ietf/SNMPv2-MIB
# Produced by pysmi-0.3.4 at Fri May  1 22:39:55 2020
# On host grumpy platform Linux version 4.15.0-88-generic by user bjk49
# Using Python version 3.6.9 (default, Apr 18 2020, 01:56:04)
#
OctetString, Integer, ObjectIdentifier = mibBuilder.importSymbols("ASN1", "OctetString", "Integer", "ObjectIdentifier")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueSizeConstraint, SingleValueConstraint, ValueRangeConstraint, ConstraintsIntersection = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueSizeConstraint", "SingleValueConstraint", "ValueRangeConstraint", "ConstraintsIntersection")
ModuleCompliance, ObjectGroup, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "ObjectGroup", "NotificationGroup")
IpAddress, snmpModules, mib_2, ModuleIdentity, Gauge32, Unsigned32, Bits, Integer32, ObjectIdentity, NotificationType, iso, Counter64, TimeTicks, Counter32, MibScalar, MibTable, MibTableRow, MibTableColumn, MibIdentifier = mibBuilder.importSymbols("SNMPv2-SMI", "IpAddress", "snmpModules", "mib-2", "ModuleIdentity", "Gauge32", "Unsigned32", "Bits", "Integer32", "ObjectIdentity", "NotificationType", "iso", "Counter64", "TimeTicks", "Counter32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "MibIdentifier")
TimeStamp, TestAndIncr, TextualConvention, DisplayString = mibBuilder.importSymbols("SNMPv2-TC", "TimeStamp", "TestAndIncr", "TextualConvention", "DisplayString")
snmpMIB = ModuleIdentity((1, 3, 6, 1, 6, 3, 1))
snmpMIB.setRevisions(('2002-10-16 00:00', '1995-11-09 00:00', '1993-04-01 00:00',))
if mibBuilder.loadTexts: snmpMIB.setLastUpdated('200210160000Z')
if mibBuilder.loadTexts: snmpMIB.setOrganization('IETF SNMPv3 Working Group')
snmpMIBObjects = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 1))
system = MibIdentifier((1, 3, 6, 1, 2, 1, 1))
sysDescr = MibScalar((1, 3, 6, 1, 2, 1, 1, 1), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysDescr.setStatus('current')
sysObjectID = MibScalar((1, 3, 6, 1, 2, 1, 1, 2), ObjectIdentifier()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysObjectID.setStatus('current')
sysUpTime = MibScalar((1, 3, 6, 1, 2, 1, 1, 3), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysUpTime.setStatus('current')
sysContact = MibScalar((1, 3, 6, 1, 2, 1, 1, 4), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 255))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sysContact.setStatus('current')
sysName = MibScalar((1, 3, 6, 1, 2, 1, 1, 5), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 255))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sysName.setStatus('current')
sysLocation = MibScalar((1, 3, 6, 1, 2, 1, 1, 6), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 255))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sysLocation.setStatus('current')
sysServices = MibScalar((1, 3, 6, 1, 2, 1, 1, 7), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 127))).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysServices.setStatus('current')
sysORLastChange = MibScalar((1, 3, 6, 1, 2, 1, 1, 8), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysORLastChange.setStatus('current')
sysORTable = MibTable((1, 3, 6, 1, 2, 1, 1, 9), )
if mibBuilder.loadTexts: sysORTable.setStatus('current')
sysOREntry = MibTableRow((1, 3, 6, 1, 2, 1, 1, 9, 1), ).setIndexNames((0, "SNMPv2-MIB", "sysORIndex"))
if mibBuilder.loadTexts: sysOREntry.setStatus('current')
sysORIndex = MibTableColumn((1, 3, 6, 1, 2, 1, 1, 9, 1, 1), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 2147483647)))
if mibBuilder.loadTexts: sysORIndex.setStatus('current')
sysORID = MibTableColumn((1, 3, 6, 1, 2, 1, 1, 9, 1, 2), ObjectIdentifier()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysORID.setStatus('current')
sysORDescr = MibTableColumn((1, 3, 6, 1, 2, 1, 1, 9, 1, 3), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysORDescr.setStatus('current')
sysORUpTime = MibTableColumn((1, 3, 6, 1, 2, 1, 1, 9, 1, 4), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysORUpTime.setStatus('current')
snmp = MibIdentifier((1, 3, 6, 1, 2, 1, 11))
snmpInPkts = MibScalar((1, 3, 6, 1, 2, 1, 11, 1), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInPkts.setStatus('current')
snmpInBadVersions = MibScalar((1, 3, 6, 1, 2, 1, 11, 3), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInBadVersions.setStatus('current')
snmpInBadCommunityNames = MibScalar((1, 3, 6, 1, 2, 1, 11, 4), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInBadCommunityNames.setStatus('current')
snmpInBadCommunityUses = MibScalar((1, 3, 6, 1, 2, 1, 11, 5), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInBadCommunityUses.setStatus('current')
snmpInASNParseErrs = MibScalar((1, 3, 6, 1, 2, 1, 11, 6), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInASNParseErrs.setStatus('current')
snmpEnableAuthenTraps = MibScalar((1, 3, 6, 1, 2, 1, 11, 30), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("enabled", 1), ("disabled", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpEnableAuthenTraps.setStatus('current')
snmpSilentDrops = MibScalar((1, 3, 6, 1, 2, 1, 11, 31), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpSilentDrops.setStatus('current')
snmpProxyDrops = MibScalar((1, 3, 6, 1, 2, 1, 11, 32), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpProxyDrops.setStatus('current')
snmpTrap = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 4))
snmpTrapOID = MibScalar((1, 3, 6, 1, 6, 3, 1, 1, 4, 1), ObjectIdentifier()).setMaxAccess("accessiblefornotify")
if mibBuilder.loadTexts: snmpTrapOID.setStatus('current')
snmpTrapEnterprise = MibScalar((1, 3, 6, 1, 6, 3, 1, 1, 4, 3), ObjectIdentifier()).setMaxAccess("accessiblefornotify")
if mibBuilder.loadTexts: snmpTrapEnterprise.setStatus('current')
snmpTraps = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 5))
coldStart = NotificationType((1, 3, 6, 1, 6, 3, 1, 1, 5, 1))
if mibBuilder.loadTexts: coldStart.setStatus('current')
warmStart = NotificationType((1, 3, 6, 1, 6, 3, 1, 1, 5, 2))
if mibBuilder.loadTexts: warmStart.setStatus('current')
authenticationFailure = NotificationType((1, 3, 6, 1, 6, 3, 1, 1, 5, 5))
if mibBuilder.loadTexts: authenticationFailure.setStatus('current')
snmpSet = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 6))
snmpSetSerialNo = MibScalar((1, 3, 6, 1, 6, 3, 1, 1, 6, 1), TestAndIncr()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpSetSerialNo.setStatus('current')
snmpMIBConformance = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 2))
snmpMIBCompliances = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 2, 1))
snmpMIBGroups = MibIdentifier((1, 3, 6, 1, 6, 3, 1, 2, 2))
snmpBasicCompliance = ModuleCompliance((1, 3, 6, 1, 6, 3, 1, 2, 1, 2)).setObjects(("SNMPv2-MIB", "snmpGroup"), ("SNMPv2-MIB", "snmpSetGroup"), ("SNMPv2-MIB", "systemGroup"), ("SNMPv2-MIB", "snmpBasicNotificationsGroup"), ("SNMPv2-MIB", "snmpCommunityGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpBasicCompliance = snmpBasicCompliance.setStatus('deprecated')
snmpBasicComplianceRev2 = ModuleCompliance((1, 3, 6, 1, 6, 3, 1, 2, 1, 3)).setObjects(("SNMPv2-MIB", "snmpGroup"), ("SNMPv2-MIB", "snmpSetGroup"), ("SNMPv2-MIB", "systemGroup"), ("SNMPv2-MIB", "snmpBasicNotificationsGroup"), ("SNMPv2-MIB", "snmpCommunityGroup"), ("SNMPv2-MIB", "snmpWarmStartNotificationGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpBasicComplianceRev2 = snmpBasicComplianceRev2.setStatus('current')
snmpGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 8)).setObjects(("SNMPv2-MIB", "snmpInPkts"), ("SNMPv2-MIB", "snmpInBadVersions"), ("SNMPv2-MIB", "snmpInASNParseErrs"), ("SNMPv2-MIB", "snmpSilentDrops"), ("SNMPv2-MIB", "snmpProxyDrops"), ("SNMPv2-MIB", "snmpEnableAuthenTraps"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpGroup = snmpGroup.setStatus('current')
snmpCommunityGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 9)).setObjects(("SNMPv2-MIB", "snmpInBadCommunityNames"), ("SNMPv2-MIB", "snmpInBadCommunityUses"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpCommunityGroup = snmpCommunityGroup.setStatus('current')
snmpSetGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 5)).setObjects(("SNMPv2-MIB", "snmpSetSerialNo"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpSetGroup = snmpSetGroup.setStatus('current')
systemGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 6)).setObjects(("SNMPv2-MIB", "sysDescr"), ("SNMPv2-MIB", "sysObjectID"), ("SNMPv2-MIB", "sysUpTime"), ("SNMPv2-MIB", "sysContact"), ("SNMPv2-MIB", "sysName"), ("SNMPv2-MIB", "sysLocation"), ("SNMPv2-MIB", "sysServices"), ("SNMPv2-MIB", "sysORLastChange"), ("SNMPv2-MIB", "sysORID"), ("SNMPv2-MIB", "sysORUpTime"), ("SNMPv2-MIB", "sysORDescr"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    systemGroup = systemGroup.setStatus('current')
snmpBasicNotificationsGroup = NotificationGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 7)).setObjects(("SNMPv2-MIB", "coldStart"), ("SNMPv2-MIB", "authenticationFailure"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpBasicNotificationsGroup = snmpBasicNotificationsGroup.setStatus('current')
snmpWarmStartNotificationGroup = NotificationGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 11)).setObjects(("SNMPv2-MIB", "warmStart"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpWarmStartNotificationGroup = snmpWarmStartNotificationGroup.setStatus('current')
snmpNotificationGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 12)).setObjects(("SNMPv2-MIB", "snmpTrapOID"), ("SNMPv2-MIB", "snmpTrapEnterprise"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpNotificationGroup = snmpNotificationGroup.setStatus('current')
snmpOutPkts = MibScalar((1, 3, 6, 1, 2, 1, 11, 2), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutPkts.setStatus('obsolete')
snmpInTooBigs = MibScalar((1, 3, 6, 1, 2, 1, 11, 8), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInTooBigs.setStatus('obsolete')
snmpInNoSuchNames = MibScalar((1, 3, 6, 1, 2, 1, 11, 9), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInNoSuchNames.setStatus('obsolete')
snmpInBadValues = MibScalar((1, 3, 6, 1, 2, 1, 11, 10), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInBadValues.setStatus('obsolete')
snmpInReadOnlys = MibScalar((1, 3, 6, 1, 2, 1, 11, 11), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInReadOnlys.setStatus('obsolete')
snmpInGenErrs = MibScalar((1, 3, 6, 1, 2, 1, 11, 12), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInGenErrs.setStatus('obsolete')
snmpInTotalReqVars = MibScalar((1, 3, 6, 1, 2, 1, 11, 13), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInTotalReqVars.setStatus('obsolete')
snmpInTotalSetVars = MibScalar((1, 3, 6, 1, 2, 1, 11, 14), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInTotalSetVars.setStatus('obsolete')
snmpInGetRequests = MibScalar((1, 3, 6, 1, 2, 1, 11, 15), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInGetRequests.setStatus('obsolete')
snmpInGetNexts = MibScalar((1, 3, 6, 1, 2, 1, 11, 16), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInGetNexts.setStatus('obsolete')
snmpInSetRequests = MibScalar((1, 3, 6, 1, 2, 1, 11, 17), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInSetRequests.setStatus('obsolete')
snmpInGetResponses = MibScalar((1, 3, 6, 1, 2, 1, 11, 18), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInGetResponses.setStatus('obsolete')
snmpInTraps = MibScalar((1, 3, 6, 1, 2, 1, 11, 19), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInTraps.setStatus('obsolete')
snmpOutTooBigs = MibScalar((1, 3, 6, 1, 2, 1, 11, 20), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutTooBigs.setStatus('obsolete')
snmpOutNoSuchNames = MibScalar((1, 3, 6, 1, 2, 1, 11, 21), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutNoSuchNames.setStatus('obsolete')
snmpOutBadValues = MibScalar((1, 3, 6, 1, 2, 1, 11, 22), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutBadValues.setStatus('obsolete')
snmpOutGenErrs = MibScalar((1, 3, 6, 1, 2, 1, 11, 24), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutGenErrs.setStatus('obsolete')
snmpOutGetRequests = MibScalar((1, 3, 6, 1, 2, 1, 11, 25), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutGetRequests.setStatus('obsolete')
snmpOutGetNexts = MibScalar((1, 3, 6, 1, 2, 1, 11, 26), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutGetNexts.setStatus('obsolete')
snmpOutSetRequests = MibScalar((1, 3, 6, 1, 2, 1, 11, 27), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutSetRequests.setStatus('obsolete')
snmpOutGetResponses = MibScalar((1, 3, 6, 1, 2, 1, 11, 28), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutGetResponses.setStatus('obsolete')
snmpOutTraps = MibScalar((1, 3, 6, 1, 2, 1, 11, 29), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpOutTraps.setStatus('obsolete')
snmpObsoleteGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 1, 2, 2, 10)).setObjects(("SNMPv2-MIB", "snmpOutPkts"), ("SNMPv2-MIB", "snmpInTooBigs"), ("SNMPv2-MIB", "snmpInNoSuchNames"), ("SNMPv2-MIB", "snmpInBadValues"), ("SNMPv2-MIB", "snmpInReadOnlys"), ("SNMPv2-MIB", "snmpInGenErrs"), ("SNMPv2-MIB", "snmpInTotalReqVars"), ("SNMPv2-MIB", "snmpInTotalSetVars"), ("SNMPv2-MIB", "snmpInGetRequests"), ("SNMPv2-MIB", "snmpInGetNexts"), ("SNMPv2-MIB", "snmpInSetRequests"), ("SNMPv2-MIB", "snmpInGetResponses"), ("SNMPv2-MIB", "snmpInTraps"), ("SNMPv2-MIB", "snmpOutTooBigs"), ("SNMPv2-MIB", "snmpOutNoSuchNames"), ("SNMPv2-MIB", "snmpOutBadValues"), ("SNMPv2-MIB", "snmpOutGenErrs"), ("SNMPv2-MIB", "snmpOutGetRequests"), ("SNMPv2-MIB", "snmpOutGetNexts"), ("SNMPv2-MIB", "snmpOutSetRequests"), ("SNMPv2-MIB", "snmpOutGetResponses"), ("SNMPv2-MIB", "snmpOutTraps"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    snmpObsoleteGroup = snmpObsoleteGroup.setStatus('obsolete')
mibBuilder.exportSymbols("SNMPv2-MIB", snmpInSetRequests=snmpInSetRequests, system=system, snmpInPkts=snmpInPkts, systemGroup=systemGroup, snmpCommunityGroup=snmpCommunityGroup, snmpOutPkts=snmpOutPkts, sysLocation=sysLocation, snmpMIB=snmpMIB, snmpInTotalReqVars=snmpInTotalReqVars, sysORUpTime=sysORUpTime, snmp=snmp, snmpSet=snmpSet, snmpObsoleteGroup=snmpObsoleteGroup, snmpOutGenErrs=snmpOutGenErrs, sysObjectID=sysObjectID, snmpOutNoSuchNames=snmpOutNoSuchNames, sysServices=sysServices, snmpOutGetResponses=snmpOutGetResponses, sysUpTime=sysUpTime, snmpOutBadValues=snmpOutBadValues, sysORID=sysORID, authenticationFailure=authenticationFailure, snmpInGenErrs=snmpInGenErrs, snmpBasicNotificationsGroup=snmpBasicNotificationsGroup, snmpMIBObjects=snmpMIBObjects, snmpTrapOID=snmpTrapOID, snmpOutSetRequests=snmpOutSetRequests, snmpInTooBigs=snmpInTooBigs, snmpOutTraps=snmpOutTraps, snmpSilentDrops=snmpSilentDrops, snmpInNoSuchNames=snmpInNoSuchNames, snmpMIBCompliances=snmpMIBCompliances, snmpInTraps=snmpInTraps, snmpNotificationGroup=snmpNotificationGroup, snmpProxyDrops=snmpProxyDrops, sysORTable=sysORTable, sysDescr=sysDescr, sysORIndex=sysORIndex, snmpInASNParseErrs=snmpInASNParseErrs, snmpInBadCommunityNames=snmpInBadCommunityNames, coldStart=coldStart, warmStart=warmStart, snmpMIBConformance=snmpMIBConformance, snmpInBadVersions=snmpInBadVersions, snmpMIBGroups=snmpMIBGroups, snmpBasicComplianceRev2=snmpBasicComplianceRev2, snmpInBadValues=snmpInBadValues, snmpSetSerialNo=snmpSetSerialNo, snmpInTotalSetVars=snmpInTotalSetVars, snmpBasicCompliance=snmpBasicCompliance, snmpOutTooBigs=snmpOutTooBigs, snmpTrapEnterprise=snmpTrapEnterprise, snmpOutGetNexts=snmpOutGetNexts, sysName=sysName, snmpSetGroup=snmpSetGroup, sysORLastChange=sysORLastChange, snmpEnableAuthenTraps=snmpEnableAuthenTraps, snmpInReadOnlys=snmpInReadOnlys, sysORDescr=sysORDescr, snmpInGetNexts=snmpInGetNexts, snmpTrap=snmpTrap, snmpInBadCommunityUses=snmpInBadCommunityUses, snmpGroup=snmpGroup, snmpWarmStartNotificationGroup=snmpWarmStartNotificationGroup, snmpInGetResponses=snmpInGetResponses, sysContact=sysContact, sysOREntry=sysOREntry, snmpOutGetRequests=snmpOutGetRequests, snmpTraps=snmpTraps, snmpInGetRequests=snmpInGetRequests, PYSNMP_MODULE_ID=snmpMIB)





class DigitalIO:
    """
    The digital IO class to read & control the digital IOs
    via the Galil actuator controller.

    Args:
        name(string)            : Name of this instance
        io_list(list)           : IO configurations
        g(gclib.py())           : Actuator controller library
        get_onoff_reverse(bool) : Return 1/0 in _get_onoff()
                                  if IO input is 0/1
        set_onoff_reverse(bool) : Send SB(1)/CB(0) to the controller
                                  in _set_onoff()
                                  if "onoff" argument is False/True
        verbose(int)    : Verbosity level
    """

    def __init__(self, name, io_list, g,
                 get_onoff_reverse=False, set_onoff_reverse=False, verbose=0):
        self.name = name
        self.g = g
        self.get_reverse = get_onoff_reverse
        self.set_reverse = set_onoff_reverse
        self.verbose = verbose

        self.io_list = io_list
        self.io_names = [io['name'] for io in io_list]
        self.io_labels = [io['label'] for io in io_list]
        self.io_indices = {io['name']: index
                           for index, io in enumerate(io_list)}
        self.io_dict = {io['name']: io['io'] for io in io_list}
        # Retrieve IO number: io OUT[3] --> 3
        self.io_numdict = \
            {name: (int)(io.split('[')[1].split(']')[0])
                for name, io in self.io_dict.items()}

    def _get_onoff(self, io_name):
        onoff = self.g.GCommand('MG @{}'.format(self.io_dict[io_name]))
        try:
            onoff = bool(float(onoff.strip()))
        except ValueError as e:
            msg = \
                'DigitalIO[{}]:_get_onoff(): ERROR!: '\
                'Failed to get correct on/off message '\
                'from the controller.\n'.format(self.name)\
                + 'DigitalIO[{}]:_get_onoff(): '\
                  ' message = "{}" | Exception = "{}"'\
                  .format(self.name, onoff, e)
            raise ValueError(msg)
        if self.get_reverse:
            onoff = not onoff
        # print('DigitalIO[{}]:_get_onoff(): onoff for {}: {}'
        #      .format(self.name, io_name, onoff))
        return int(onoff)

    def get_onoff(self, io_name=None):
        """Get True/False (ON/OFF) for the digital IOs.

        Args:
            io_name (str or list): A string of a IO name or list of IO names.

        Returns:
            list or str: A list of True/Falses or a single True/False depending
            on the value of `io_name`. If `io_name` is None, return a list of
            the ON/OFFs for all the IOs. If `io_name` is a list, return a list
            of the ON/OFFs for asked IOs.  If `io_name` is a string (one IO),
            return one ON/OFF. Here True means ON and False means OFF.
        """
        if io_name is None:
            onoff = [self._get_onoff(name) for name in self.io_names]
        elif isinstance(io_name, list):
            if not all([(name in self.io_names) for name in io_name]):
                msg = \
                    'DigitalIO[{}]:get_onoff(): ERROR!: '\
                    .format(self.name) \
                    + 'There is no matched IO name.\n'\
                      'DigitalIO[{}]:get_onoff():       '\
                      .format(self.name)\
                    + 'Assigned IO names = {}\n'\
                      .format(self.io_names)\
                    + 'DigitalIO[{}]:get_onoff():       '\
                      'Asked IO names = {}'.format(self.name, io_name)
                raise ValueError(msg)
            onoff = [self._get_onoff(name) for name in io_name]
        else:
            if not (io_name in self.io_names):
                msg = \
                    'DigitalIO[{}]:get_onoff(): ERROR!: '\
                    'There is no IO name of {}.\n'\
                    .format(self.name, io_name) \
                    + 'DigitalIO[{}]:get_onoff():         '\
                      'Assigned IO names = {}'.format(self.name, self.io_names)
                raise ValueError(msg)
            onoff = self._get_onoff(io_name)
        return onoff

    def get_label(self, io_name):
        if io_name is None:
            label = self.io_labels
        elif isinstance(io_name, list):
            if not all([(name in self.io_names) for name in io_name]):
                msg = \
                    'DigitalIO[{}]:get_label(): ERROR!: '\
                    'There is no matched IO name.\n'\
                    .format(self.name)\
                    + 'DigitalIO[{}]:get_label():     '\
                      'Assigned IO names = {}\n'\
                      .format(self.name, self.io_names)\
                    + 'DigitalIO[{}]:get_label():     '\
                      'Asked IO names    = {}'.format(self.name, io_name)
                raise ValueError(msg)
            label = [self.io_indices[name] for name in io_name]
        else:
            if not (io_name in self.io_names):
                msg = \
                    'DigitalIO[{}]:get_label(): ERROR!: '\
                    'There is no IO name of {}.\n'\
                    .format(self.name, io_name) \
                    + 'DigitalIO[{}]:get_label():     '\
                      'Assigned IO names = {}'.format(self.name, self.io_names)
                raise ValueError(msg)
            label = self.io_label(io_name)
        return label

    def _set_onoff(self, onoff, io_name):
        io_num = self.io_numdict[io_name]
        onoff = bool(int(onoff))
        if self.set_reverse:
            onoff = not onoff
        if onoff:
            cmd = 'SB {}'.format(io_num)
        else:
            cmd = 'CB {}'.format(io_num)
        self.g.GCommand(cmd)

    def set_onoff(self, onoff=0, io_name=None):
        """Set True/False (ON/OFF) for the digital IOs.

        Args:
            onoff (int): 0 (OFF) or 1 (ON)
            io_name (str or list): a string of a IO name or list of IO names.
                If None, set all possible IOs ON/OFF. If a list, set all
                specified IOs ON/OFF. If a string (one IO), set the given IO
                ON/OFF.
        """
        set_io_names = []
        if io_name is None:
            set_io_names = self.io_names
        elif isinstance(io_name, list):
            if not all([(name in self.io_names) for name in io_name]):
                msg = \
                    'DigitalIO[{}]:set_onoff(): ERROR!: '\
                    'There is no matched IO name.\n'\
                    .format(self.name)\
                    + 'DigitalIO[{}]:set_onoff():    '\
                      'Assigned IO names = {}\n'\
                      .format(self.name, self.io_names)\
                    + 'DigitalIO[{}]:set_onoff():     '\
                      'Asked IO names    = {}'\
                      .format(self.name, io_name)
                raise ValueError(msg)
            set_io_names = io_name
        else:
            if not (io_name in self.io_names):
                msg = \
                    'DigitalIO[{}]:set_onoff(): ERROR!: '\
                    'There is no IO name of {}.\n'\
                    .format(self.name, io_name)\
                    + 'DigitalIO[{}]:set_onoff():     '\
                      'Assigned IO names = {}'.format(self.name, self.io_names)
                raise ValueError(msg)
            set_io_names = [io_name]
        print('DigitalIO[{}]:set_onoff(): Set {} for the IOs: '
              '{}'.format(self.name, 'ON' if onoff else 'OFF', set_io_names))
        for name in set_io_names:
            self._set_onoff(onoff, name)

    def set_allon(self):
        print('DigitalIO[{}]:set_allon(): '
              'Set ON for all of the digital IOs'.format(self.name))
        self.set_onoff(1, None)

    def set_alloff(self):
        print('DigitalIO[{}]:set_allon(): '
              'Set OFF for all of the digital IOs'.format(self.name))
        self.set_onoff(0, None)



import argparse
import os
import time
from typing import Optional

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.vantagepro2.drivers import VantagePro2


class VantagePro2Agent:
    """Agent to connect to single VantagePro2 Weather Monitor Device.

    Args:
        sample_freq (double):
             frequency (Hz) at which the weather monitor samples data. Can not
             be faster than 0.5 Hz. This value is converted to period (sec) for
             time.wait(seconds)
        port (string):
             usb port that the weather monitor is connected to.  The monitor
             will communicate via this port.
    """
    # change port argument when I figure out how to generalize it!

    def __init__(self, agent, port="/dev/ttyUSB0", sample_freq=0.5):
        self.active = True
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.port = port
        self.module: Optional[VantagePro2] = None

        if sample_freq > 0.5:
            self.log.warn("Sample frequency too fast! Setting to 0.5Hz")
            sample_freq = 0.5
        self.sample_freq = sample_freq

        self.initialized = False
        self.take_data = False

        # Registers weather data feed
        agg_params = {
            'frame_length': 60,
        }
        self.agent.register_feed('weather_data',
                                 record=True,
                                 agg_params=agg_params)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init(self, session, params=None):
        """init(auto_acquire=False)

        **Task** - Perform first time setup of the Weather Monitor Module.

        Parameters:
            auto_acquire (bool): Automatically start acq process after
                initialization if True. Defaults to False.

        """
        if self.initialized:
            return True, "Already Initialized Module"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('starting')

            self.module = VantagePro2(self.port)
            print("Initialized Vantage Pro2 module: {!s}".format(
                self.module))

        self.initialized = True

        # Start data acquisition if requested
        if params['auto_acquire']:
            self.agent.start('acq')

        time.sleep(2)

        return True, 'Vantage Pro2 module initialized.'

    @ocs_agent.param('sample_freq', default=0.5, type=float)
    def acq(self, session, params=None):
        """acq(sample_freq=0.5)

        **Process** - Start data acquisition.

        Parameters:
            sample_freq (float):
                Frequency at which weather data is sampled. Defaults to 0.5
                Hz.

        """
        sample_freq = params['sample_freq']
        # If loops is None, use value passed to Agent init
        if sample_freq is None:
            sample_freq = self.sample_freq
        wait_time = 1 / sample_freq

        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("""Could not start acq because {} is
                already running"""
                              .format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')
            # use pacemaker class to take data at regular intervals
            if sample_freq % 1 == 0:
                pm = Pacemaker(sample_freq, True)
            else:
                pm = Pacemaker(sample_freq)

            self.take_data = True

            while self.take_data:
                pm.sleep()
                data = {
                    'timestamp': time.time(),
                    'block_name': 'weather',
                    'data': {}
                }
                data['data'] = self.module.weather_daq()
                self.agent.publish_to_feed('weather_data', data)
                time.sleep(wait_time)

            self.agent.feeds['weather_data'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """

        if self.take_data:
            self.take_data = False
            print('requested to stop taking data.')
            return True, 'data taking succesfully halted'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', type=str,
                        help="Path to USB node for the Vantage Pro2")
    pgroup.add_argument('--serial-number', type=str,
                        help="Serial number of VantagePro2 Monitor")
    pgroup.add_argument('--mode', type=str, choices=['idle', 'init', 'acq'],
                        help="Starting action for the agent.")
    pgroup.add_argument('--sample-freq', type=float,
                        help="Sample frequency for weather data collection")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='VantagePro2Agent',
                                  parser=parser,
                                  args=args)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    device_port = None

    if args.port is not None:
        device_port = args.port
    else:  # Tries to find correct USB port automatically

        # This exists if udev rules are setup properly for the 240s
        if os.path.exists('/dev/{}'.format(args.serial_number)):
            device_port = "/dev/{}".format(args.serial_number)

        elif os.path.exists('/dev/serial/by-id'):
            ports = os.listdir('/dev/serial/by-id')
            for port in ports:
                if args.serial_number in port:
                    device_port = "/dev/serial/by-id/{}".format(port)
                    print("Found port {}".format(device_port))
                    break

    if device_port is None:
        print("Could not find device port for {}".format(args.serial_number))
        return

    agent, runner = ocs_agent.init_site_agent(args)

    vPro2 = VantagePro2Agent(agent, device_port, args.sample_freq)
    agent.register_task('init', vPro2.init,
                        startup=init_params)
    agent.register_process('acq', vPro2.acq, vPro2._stop_acq,
                           blocking=True)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



# Digital IO
IO_INFO = [
    # Right stopper: Inside of the forebaffle interface
    {'name': 'STR1', 'io': 'OUT[1]', 'label': 'Right1Inside'},
    # Right stopper: Outside of the forebaffle interface
    {'name': 'STR2', 'io': 'OUT[2]', 'label': 'Right2Outside'},
    # Left stopper: Inside of the forebaffle interface
    {'name': 'STL1', 'io': 'OUT[3]', 'label': 'Left1Inside'},
    # Left stopper: Outside of the forebaffle interface
    {'name': 'STL2', 'io': 'OUT[4]', 'label': 'Left2Outside'},
]



import argparse
import subprocess
import time

import numpy as np
import txaio

txaio.use_twisted()

from ocs import ocs_agent, site_config


def get_sensors(shm_addr):
    """
    Runs a command on the shelf manager that returns a list of all
    of the avialable sensors to stdout. Uses subprocess module to
    read stdout and identify the ipmb address and sensor id for all
    sensors which are Threshold type as opposed to discrete type,
    which are alarms.
    Args:
        shm_addr (str):
            Address used to connect to shelf manager ex. root@192.168.1.2
    Returns:
        ipmbs (str list):
            List of Intelligent Platform Management Bus (IPMB) addresses
        sensids (str list):
            List of sensor identification names, same length as ipmbs list.
    """
    log = txaio.make_logger()

    # SSH to shelf manager
    cmd = ['ssh', f'{shm_addr}']
    # Send command to shelf manager
    cmd += ['clia', 'sensordata']
    # Intialize output data
    ipmbs = []
    sensids = []
    masksens = []
    check_sense = False

    # Send command to ssh and run command on shelf
    ssh = subprocess.Popen(cmd,
                           shell=False,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
    # Readback shelfmanager standard out
    result = ssh.stdout.readlines()
    # Parse readback data line by line unless empty
    if result == []:
        error = ssh.stderr.readlines()
        log.error("ERROR: %s" % error)
    else:
        for r in result:
            if ': LUN' in r.decode('utf-8'):
                check_sense = True
                ipmbs.append(r.decode('utf-8').split(': LUN')[0])
                sname = r.decode('utf-8').split('(')[-1].split(')')[0]
                sensids.append(sname)
                continue
            if check_sense:
                if 'Threshold' in r.decode('utf-8'):
                    masksens.append(True)
                if 'Discrete' in r.decode('utf-8'):
                    masksens.append(False)
                check_sense = False
    ipmbs = np.asarray(ipmbs)
    sensids = np.asarray(sensids)
    masksens = np.asarray(masksens)
    return ipmbs[masksens], sensids[masksens]


def get_channel_names(ipmbs):
    """
    Converts ipmb addresses to human readable names based on the
    definitions of ipmb addresses in the ATCA manuals.
    Args:
        ipmbs (str list):
            List of Intelligent Platform Management Bus (IPMB) addresses
    Returns:
        chan_names (str list):
            List of human readable names for each IPMB address.
    """
    chan_names = np.zeros(len(ipmbs)).astype(str)
    for i, ipmb in enumerate(ipmbs):
        if ipmb == '20':
            chan_names[i] = 'shelf'
            continue
        if ipmb == 'fe':
            chan_names[i] = 'pwr_mgmt'
            continue
        slot = int('0x' + ipmb, 16) // 2 - 64
        if slot < 1:
            # Not exactly sure what this corresponds to...
            chan_names[i] = f"ipmb{ipmb}"
            continue
        if slot == 1:
            chan_names[i] = 'switch'
            continue
        chan_names[i] = f'slot{slot}'
    return chan_names


def get_data_dict(shm_addr, ipmbs, sensids, chan_names,
                  crate_id):
    """
    Given a list of ipmb addresses, sensor ids, and channel names,
    the shelf manager is queeried and the current sensor values for
    the provided list of sensors is read. The values are then
    output in a dictionary in the format needed to publish to
    influxdb.
    Args:
        shm_addr (str):
            Address used to connect to shelf manager ex. root@192.168.1.2
        ipmbs (str list):
            List of Intelligent Platform Management Bus (IPMB) addresses.
        sensids (str list):
            List of sensor identification names, same length as ipmbs list.
        chan_names (str list):
            List of human readable names for each IPMB address.
        crate_id (str):
            String to identify crate number in feed names, ex: crate_1
    Returns:
        data_dict (dict):
            Dict with structure, {data : value} collects the output
            of all of the sensors passed into the fuction. Ensures the
            keys match the influxdb feedname requirements
    """
    log = txaio.make_logger()

    data_dict = {}
    cmd = ['ssh', f'{shm_addr}', 'clia', 'sensordata']
    ssh = subprocess.Popen(cmd,
                           shell=False,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
    result = ssh.stdout.readlines()
    if result == []:
        error = ssh.stderr.readlines()
        log.error("ERROR: %s" % error)
    else:
        for ipmb, sensid, chan_name in zip(ipmbs, sensids, chan_names):
            sense_chan = False
            for r in result:
                if ipmb in r.decode('utf-8'):
                    if sensid in r.decode('utf-8'):
                        sense_chan = True
                        continue
                if sense_chan:
                    if 'Processed data:' in r.decode('utf-8'):
                        sid = sensid.strip('"')
                        sid = sid.replace(" ", "_")
                        sid = sid.replace(":", "")
                        sid = sid.replace("+", "")
                        sid = sid.replace(".", "p")
                        sid = sid.replace("-", "_")
                        line = r.strip().decode("utf-8")
                        if line.split(':')[-1].split(' ')[0] == '':
                            val = float(line.split(':')[-1].split(' ')[1])
                        else:
                            val = float(line.split(':')[-1].split(' ')[0])
                        data_dict[f'{crate_id}_{chan_name}_{sid}'] = val
                        sense_chan = False
    return data_dict


class SmurfCrateMonitor:
    def __init__(self, agent, crate_id, shm_addr):
        self.agent = agent
        self.log = agent.log
        self.shm_addr = shm_addr
        self.crate_id = crate_id
        # Register feed
        agg_params = {
            'frame_length': 10 * 60
        }
        self.log.info('registering')
        self.agent.register_feed('smurf_sensors',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0.)

    def _init_data_stream(self, shm_addr):
        """Wrapper for get_sensors and get_channel_names which generates the
        list of sensors to use in datastreaming.

        Args:
            shm_addr (str): Address used to connect to shelf manager ex.
                root@192.168.1.2
        Return:
            ipmbs (str list): List of Intelligent Platform Management Bus
                (IPMB) addresses.
            sensids (str list): List of sensor identification names, same
                length as ipmbs list.
            chan_names (str list): List of human readable names for each IPMB
                address.
        """
        ipmbs, sensids = get_sensors(shm_addr)
        chan_names = get_channel_names(ipmbs)
        return ipmbs, sensids, chan_names

    def init_crate(self, session, params=None):
        """init_crate()

        **Task** - Initialize connection to the SMuRF crate.

        Run at the startup of the docker to check that you can
        successfully ssh to the crate and run a command. If it runs
        successfully then you should see the home directory of the shelf
        manager printed to the docker logs and the data acquisition process to
        start, if not you will see an error in the logs and acquistion won't
        start.

        """
        self.log.info(self.shm_addr)
        cmd = ['ssh', f'{self.shm_addr}', 'pwd']
        self.log.info("command run: {c}", c=cmd)
        ssh = subprocess.Popen(cmd,
                               shell=False,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
        result = ssh.stdout.readlines()
        self.log.info(result[0])
        if result == []:
            error = ssh.stderr.readlines()
            self.log.error(f"ERROR: {error}")
            return False, 'Crate failed to initialize'
        if result[0].decode("utf-8") == '/etc/home/root\n':
            self.log.info('Successfully ssh-d into shelf')
            self.agent.start('acq')
            return True, 'Crate Initialized'

    def acq(self, session, params=None):
        """acq()

        **Process** - Start acquiring data.

        Hardcoded for one data point every 30 seconds because we intend for
        this to be very low rate data.

        """
        self.log.info('Started acquisition')
        shm_addr = self.shm_addr
        ipmbs, sensids, chan_names = self._init_data_stream(shm_addr=shm_addr)
        self.log.info('Got sensor names')
        self.take_data = True
        while self.take_data:
            for _ in range(30):
                if not self.take_data:
                    break
                time.sleep(1)
            datadict = get_data_dict(shm_addr=self.shm_addr,
                                     ipmbs=ipmbs,
                                     sensids=sensids,
                                     chan_names=chan_names,
                                     crate_id=self.crate_id)
            data = {
                'timestamp': time.time(),
                'block_name': f'smurf_{self.crate_id}',
                'data': datadict
            }
            self.agent.publish_to_feed('smurf_sensors', data)
        return True, 'Acquisition exited cleanly'

    def _stop_acq(self, session, params=None):
        """
        Stops acquiring data if the dpcler os stopped.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()
    # Add options specific to this agent.
    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument('--shm-addr',
                        help='Shelf manager addres i.e. root@192.168.1.2')
    pgroup.add_argument('--crate-id',
                        help='Crate id used for block_name')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='CrateAgent',
                                  parser=parser,
                                  args=args)
    startup = True
    agent, runner = ocs_agent.init_site_agent(args)
    shm_addr = args.shm_addr
    crate_id = args.crate_id

    smurfcrate = SmurfCrateMonitor(agent, crate_id, shm_addr)

    agent.register_task('init_crate', smurfcrate.init_crate,
                        startup=startup)
    agent.register_process('acq', smurfcrate.acq,
                           smurfcrate._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import ast
import os
import queue
import time

import numpy as np
import so3g  # noqa: F401
import txaio
import yaml
from ocs import ocs_agent, site_config
from scipy import signal
from spt3g import core

MAX_CHANS = 4096
CHANS_PER_BAND = 512
pA_per_rad = 9e6 / (2 * np.pi)


# Map from primary key-names to their index in the SuperTimestream
# This will be populated when the first frame comes in
primary_idxs = {}


def load_frame_data(frame):
    """
    Returns detector data from a G3Stream.

    Returns:
        times : np.ndarray
            Array with shape (nsamps) of timestamps (sec)
        data : np.ndarray
            Array with shape (nchans, nsamps) of detector phase data (phi0)
    """
    primary = frame['primary']
    if isinstance(primary, core.G3TimesampleMap):
        times = np.array(primary['UnixTime']) / 1e9
    else:
        if not primary_idxs:
            for i, name in enumerate(frame['primary'].names):
                primary_idxs[name] = i
        times = np.array(primary.data[primary_idxs['UnixTime']]) / 1e9

    d = frame['data']
    if isinstance(d, core.G3TimestreamMap):
        nchans, nsamps = len(d), d.n_samples
        data = np.ndarray((nchans, nsamps), dtype=np.float32)
        for i in range(nchans):
            data[i] = d[f'r{i:0>4}'] * (2 * np.pi) / 2**16

    else:  # G3SuperTimestream probably
        data = d.data * (2 * np.pi) / 2**16

    return times, data


def sleep_while_running(duration, session, interval=1):
    """
    Sleeps for a certain duration as long as a session object's status is
    'starting' or 'running'. If the session is changed to 'stopping',
    this will quickly return False. If the sleep is completed without
    interruption this will return True.

    Args
    ----
    duration : float
        Amount of time (sec) to sleep for.
    session : OpSessions
        Session whose status should be monitored
    interval : float, optional
        Polling interval (sec) to check the session status. Defaults to 1 sec.
    """
    end_time = time.time() + duration
    while session.status in ['starting', 'running']:
        now = time.time()
        if now >= end_time:
            return True
        time.sleep(min(interval, end_time - now))
    return False


class FIRFilter:
    """
    Class for Finite Input Response filter. Filter state is preserved
    between `lfilt` calls so you can filter frame-based data.

    From scipy docs, the output of the filter is determined by:

        a[0]*y[n] = b[0]*x[n] + b[1]*x[n-1] + ... + b[M]*x[n-M]
                              - a[1]*y[n-1] - ... - a[N]*y[n-N]
    """

    def __init__(self, b, a, nchans=None):
        if nchans is None:
            nchans = MAX_CHANS
        self.b = b
        self.a = a
        self.z = np.zeros((nchans, len(b) - 1))

    def lfilt(self, data, in_place=True):
        """Filters data in place"""
        n = len(data)
        if in_place:
            data[:, :], self.z[:n] = signal.lfilter(
                self.b, self.a, data, axis=1, zi=self.z[:n]
            )
        else:
            d, self.z[:n] = signal.lfilter(
                self.b, self.a, data, axis=1, zi=self.z[:n]
            )
            return d

    @classmethod
    def butter_highpass(cls, cutoff, fs, order=5):
        """
        Creates an highpass butterworth FIR filter

        Args
        ----
        cutoff : float
            Cutoff freq (Hz)
        fs : float
            sample frequency (Hz)
        order : int
            Order of the filter
        """
        nyq = 0.5 * fs
        normal_cutoff = cutoff / nyq
        b, a = signal.butter(order, normal_cutoff, btype='high', analog=False)
        return cls(b, a)

    @classmethod
    def butter_lowpass(cls, cutoff, fs, order=5):
        """
        Creates an lowpass butterworth FIR filter

        Args
        ----
        cutoff : float
            Cutoff freq (Hz)
        fs : float
            sample frequency (Hz)
        order : int
            Order of the filter
        """
        nyq = 0.5 * fs
        normal_cutoff = cutoff / nyq
        b, a = signal.butter(order, normal_cutoff, btype='low', analog=False)
        return cls(b, a)

    @classmethod
    def moving_avg(cls, width):
        """
        Creates a moving avg filter

        Args
        -----
        width : int
            number of samples to average together
        """
        b = 1. / width * np.ones(width, dtype=float)
        a = np.zeros_like(b)
        a[0] = 1
        return cls(b, a)

    @classmethod
    def differ(cls, delay=1):
        """
        FIR filter to calculate a rolling diff of incoming data.

        Args
        ----
        delay : int
            Sets the diff spacing
        """
        b = np.zeros(1 + int(delay))
        a = np.zeros_like(b)
        b[0], b[-1] = 1, -1
        a[0] = 1
        return cls(b, a)


class Demodulator:
    """
    Helper class for demodulating a live timestream

    Args
    -----
    f : float
        Demodulation frequency
    bw : float
        Bandwidth. This will be the filter-cutoff of the applied lowpass filter
    fs : float
        Sample rate of incoming data
    """

    def __init__(self, f, bw=1, fs=200):
        self.f = f
        self.lp_sin = FIRFilter.butter_lowpass(bw, fs)
        self.lp_cos = FIRFilter.butter_lowpass(bw, fs)

    def apply(self, times, data):
        """
        Applies demodulation to data segment.

        Returns
        --------
        demod : np.ndarray
            Array of (unnormalized) demodulated data in the same shape as the
            input data
        """
        sin = np.sin(2 * np.pi * self.f * times)
        cos = np.cos(2 * np.pi * self.f * times)

        # We don't really care about normalization
        demod_sin = self.lp_sin.lfilt(data * sin[None, :], in_place=False)
        demod_cos = self.lp_cos.lfilt(data * cos[None, :], in_place=False)

        return np.sqrt(demod_sin**2 + demod_cos**2)


class WhiteNoiseCalculator:
    """
    Helper class for calculating white noise levels of incoming data

    Args
    -----
    fs : float
        Sample rate of incoming data
    navg : int
        Number of samples to average over in the RMS calc
    """

    def __init__(self, fs=200, navg=200):
        # Aiming for 20 Hz
        delay = fs // 20
        self.differ = FIRFilter.differ(delay=delay)
        self.averager = FIRFilter.moving_avg(navg)
        self.fsamp = fs

    def apply(self, data):
        """
        Returns rms / sqrt(fsamp), which estimates the white noise level
        """
        return np.sqrt(
            self.averager.lfilt(
                self.differ.lfilt(data, in_place=False)**2, in_place=False
            ) / self.fsamp
        )


class VisElem:
    """
    Container for config info for Lyrebird visual elements.

    Attributes
    --------------
    name : str
        Name of the channel
    x : float
        x-coord of the element
    y : float
        y-coord of the element
    rot : float
        Rotation angle of the element (rads)
    vals : List[str]
        List containing the names of the data values corresponding to this
        visual element. All vis-elems must have the same number of
        data-values. Data-values must be unique, such as
        ``<channel_name>/rms``
    eqs : List[str]
        List of equations to be displayed by the visual element. Equations
        are written in Polish Notation, and may contain a combination of
        numbers and data-values. Data-values can be global as defined in
        the lyrebird config file, or channel values as defined in the
        ``value_names`` argument. There must be the same number of eqs per
        visual element.

        Polish notation is a method of writing equations where operators
        precede the operands, making it easier to parse and evaluate. For
        example, the operation :math:`a + b` will be ``+ a b`` in polish
        notation, and :math:`(a + b) / 2` can be written as ``/ + a b 2``.
        See the magpie docs page for a full list of operators that are
        accepted by lyrebird.
    eq_labels : List[str]
        List of strings used to label equations in the lyrebird gui. This
        list must have the same size as the ``eqs`` array.
    cmaps : List[str]
        List of colormaps to use for each equation. This must be the same
        size as the ``eqs`` array.
    template : str
        Template used for this vis-elem. Templates are defined in the
        lyrebird cfg file.
    abs_smurf_chan : int
        Absolute smurf-channel corresponding to this visual element.
    eq_color_is_dynamic : List[bool]
        List of booleans that determine if each equation's color-scale is
        dynamic or fixed. This must be the same size as the ``eqs`` array.
    """

    def __init__(self, name, x, y, rot, template, abs_smurf_chan, cmap_idx=0):
        self.x = x
        self.y = y
        self.rot = rot
        self.template = template
        self.abs_smurf_chan = abs_smurf_chan
        self.name = name

        vals = [
            '{name}/raw', '{name}/demod', '{name}/wl', '{name}/flagged',
            '{name}/smurf_band', '{name}/smurf_chan'
        ]
        self.vals = [v.format(name=name) for v in vals]
        eq_templates = {
            'raw': '{name}/raw',
            'demod': '* {name}/demod rms_scale',
            'wl': '* {name}/wl wl_scale',
            'flagged': '{name}/flagged',
            'smurf_band': f'{self.abs_smurf_chan // CHANS_PER_BAND}',
            'smurf_chan': f'{self.abs_smurf_chan % 512}',
        }

        self.eqs = [
            eq.format(name=name) for eq in eq_templates.values()
        ]
        self.eq_labels = [k for k in eq_templates.keys()]
        self.color_is_dynamic = [False for _ in self.eqs]
        self.color_is_dynamic[0] = True

        self.cmaps = [
            ['red_cmap' for _ in eq_templates],
            ['blue_cmap' for _ in eq_templates]
        ][cmap_idx]


class FocalplaneConfig:
    def __init__(self):
        """
        Object to configure the focal-plane layout.

        Attributes
        -------------
        chan_mask : np.ndarray
            Map from absolute_smurf_chan --> Visual Element index
        channels : list
            List of visual elements
        """
        self.channels = []
        self.chan_mask = np.full(MAX_CHANS, -1)

    def config_frame(self):
        """
        Generates a config frame for lyrebird
        """
        xs = []
        ys = []
        rots = []
        cnames = []
        templates = []
        value_names = []
        eqs = []
        eq_labels = []
        eq_color_is_dynamic = []
        cmaps = []

        for c in self.channels:
            xs.append(c.x)
            ys.append(c.y)
            rots.append(c.rot)
            cnames.append(c.name)
            templates.append(c.template)
            value_names.extend(c.vals)
            eqs.extend(c.eqs)
            eq_labels.extend(c.eq_labels)
            eq_color_is_dynamic.extend(c.color_is_dynamic)
            cmaps.extend(c.cmaps)

        frame = core.G3Frame(core.G3FrameType.Wiring)
        frame['x'] = core.G3VectorDouble(xs)
        frame['y'] = core.G3VectorDouble(ys)
        frame['cname'] = core.G3VectorString(cnames)
        frame['rotation'] = core.G3VectorDouble(rots)
        frame['templates'] = core.G3VectorString(templates)
        frame['values'] = core.G3VectorString(value_names)
        frame['color_is_dynamic'] = core.G3VectorBool(eq_color_is_dynamic)
        frame['equations'] = core.G3VectorString(eqs)
        frame['eq_labels'] = core.G3VectorString(eq_labels)
        frame['cmaps'] = core.G3VectorString(cmaps)
        return frame

    def add_vis_elem(self, *args, **kwargs):
        """
        Adds a visual element to the focal-plane and updates the channel mask
        """
        c = VisElem(*args, **kwargs)
        self.channels.append(c)
        self.chan_mask[c.abs_smurf_chan] = len(self.channels) - 1

    @classmethod
    def grid(cls, stream_id, xdim, ydim, ygap=0, offset=(0., 0.)):
        """
        Creates a FocalplaneConfig object for a grid of channels.

        Args
        ----
        stream_id : str
            Stream-id for the magpie agent. This will be prepended to all
            lyrebird data-val names.
        xdim : int
            Number of channels in the x-dim of the grid
        ydim : int
            Number of channels in the y-dim of the grid
        ygap : int
            A small gap will be added every ``ygap`` rows, to better organize
            channels.
        offset : Tuple(float, float)
            Global offset of the grid with respect to the lyrebird
            coordinate-system
        """
        fp = cls()
        xs, ys = np.arange(xdim), np.arange(ydim)
        xs = xs + offset[0]
        ys = ys + offset[1]

        # Adds gaps every ygap rows
        if ygap > 0:
            ys = ys + .5 * (ys // ygap)

        template = 'box'
        for i in range(xdim * ydim):
            x, y = xs[i % xdim], ys[i // xdim]
            name = f"{stream_id}/channel_{i}"
            fp.add_vis_elem(name, x, y, 0, template, i)

        return fp

    @classmethod
    def from_csv(cls, stream_id, detmap_file, wafer_scale=1., offset=(0, 0)):
        """
        Creates a FocalplaneConfig object from a detmap csv file.

        Args
        -----
        stream_id : str
            Stream-id for the magpie agent. This will be prepended to all
            lyrebird data-val names.
        detmap_file : str
            Path to detmap csv file.
        wafer_scale : int
            Scalar to multiply against det x and y positions when translating
            to lyrebird positions. Defaults to 1, meaning that the lyrebird
            coordinate system will be the same as the det-map coordinate system,
            so x and y will be in um.
        offset : Tuple(float, float)
            Global offset of the grid with respect to the lyrebird
            coordinate-system. If wafer_scale is 1, this should be in um.
        """
        import pandas as pd
        fp = cls()
        df = pd.read_csv(detmap_file)

        templates = ["template_c0_p0", "template_c1_p0", ]

        color_idxs = {}
        ncolors = 0
        for i, row in df.iterrows():
            rot = 0
            try:
                bandpass = int(row['bandpass'])
                if bandpass in color_idxs:
                    cidx = color_idxs[bandpass]
                else:
                    color_idxs[bandpass] = ncolors
                    ncolors += 1
                    cidx = color_idxs[bandpass]

                template = templates[cidx]
                if row['pol'].strip() == 'B':
                    rot = np.pi / 2
            except ValueError:
                # Just skip detectors with unknown bandpass
                continue
            x, y = row['det_x'] * wafer_scale, row['det_y'] * wafer_scale
            x += offset[0]
            y += offset[1]

            band, chan = row['smurf_band'], row['smurf_channel']
            if chan == -1:
                continue
            abs_smurf_chan = band * CHANS_PER_BAND + chan

            name = f"{stream_id}/det_{i}"
            fp.add_vis_elem(name, x, y, rot, template, abs_smurf_chan, cmap_idx=cidx)

        return fp


class MagpieAgent:
    """
    Agent for processing streamed G3Frames, and sending data to lyrebird.

    Attributes
    -----------
    target_rate : float
        This is the target sample rate of data to be sent to lyrebird.
        Incoming data will be downsampled to this rate before being sent out.
    ds_offset : int
        Offset for downsample to avoid hiccups at frame boundaries
    fp : FocalplaneConfig
        This is the FocalplaneConfig object that contains info about what
        channels are present in the focal-plane representation, and their
        positions.
    mask : np.ndarray
        This is a channel mask that maps readout channel to
        absolute-smurf-chan. Before a status frame containing the ChannelMask
        is seen in the G3Stream, this defaults to being an identity mapping
        which just sends the readout channel no. to itself. Once a status
        frame with the channel mask is seen, this will be updated
    out_queue : Queue
        This is a queue containing outgoing G3Frames to be sent to lyrebird.
    delay : float
        The outgoing stream will attempt to enforce this delay between the
        relative timestamps in the G3Frames and the real time to ensure a
        smooth flow of data. This must be greater than the frame-aggregation
        time of the SmurfStreamer or else lyrebird will update data in spurts.
    avg1, avg2 : RollingAvg
        Two Rolling Averagers which are used to calculate the rolling RMS data.
    monitored_channels : list
        List of monitored channels whose data should be sent to grafana.
        This list will contain entries which look like
        ``(readout_chan_number, field_name)``.
    monitored_chan_sample_rate : float
        Sample rate (Hz) to target when downsampling monitored channel data for
        grafana.
    demod : Demodulator
        Demodulator used to calculate demod signal for incoming timestreams
    self.demod_freq : float
        Demodulation frequency
    self.demod_bandwidth : float
        Filter cutoff for demodulation lowpass
    wlcalc : WhiteNoiseCalculator
        WhiteNoiseCalculator used to calculate white noise levels for incoming
        timestreams
    """
    mask_register = 'AMCc.SmurfProcessor.ChannelMapper.Mask'

    def __init__(self, agent, args):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = self.agent.log
        self._running = False

        self.target_rate = args.target_rate
        self.ds_offset = 0
        layout = args.layout.lower()
        if layout == 'grid':
            self.fp = FocalplaneConfig.grid(
                args.stream_id, args.xdim, args.ydim, ygap=8, offset=args.offset
            )
        elif layout == 'wafer':
            if args.det_map is not None:
                self.fp = FocalplaneConfig.from_csv(
                    args.stream_id, args.det_map, wafer_scale=args.wafer_scale, offset=args.offset
                )
            else:
                raise ValueError("CSV file must be set using the det-map arg if "
                                 "using wafer layout")

        self.mask = np.arange(MAX_CHANS)
        self.out_queue = queue.Queue(1000)
        self.delay = args.delay

        self.demod = None
        self.demod_freq = args.demod_freq
        self.demod_bandwidth = args.demod_bandwidth

        self.wncalc = None

        self.monitored_channels = []
        self.monitored_chan_sample_rate = 10
        self.agent.register_feed(
            'detector_tods', record=True,
            agg_params={'exclude_aggregator': True}
        )
        self.agent.register_feed('white_noise', record=True,)

    @ocs_agent.param('target_rate', type=float)
    def set_target_rate(self, session, params):
        """set_target_rate(target_rate)

        Sets the target downsampled sample-rate of the data sent to lyrebird

        Args:
            target_rate : float
                Target sample rate for lyrebird (Hz)
        """
        self.target_rate = params['target_rate']
        return True, f'Set target rate to {self.target_rate}'

    @ocs_agent.param('delay', type=float)
    def set_delay(self, session, params):
        """set_delay(delay)

        Sets the target downsampled sample-rate of the data sent to lyrebird

        Args:
            target_rate : float
                Target sample rate for lyrebird (Hz)
        """
        self.delay = params['delay']
        return True, f'Set delay param to {self.delay}'

    @ocs_agent.param('chan_info', type=list, check=lambda x: len(x) <= 6)
    @ocs_agent.param('sample_rate', type=float, default=10,
                     check=lambda x: 0 < x <= 20)
    def set_monitored_channels(self, session, params):
        """set_monitored_channels(chan_info, sample_rate=10)

        **Task** - Sets channels which will be monitored and have their
        downsampled data sent to grafana.

        Field names can be manually specified in the chan_info list because it
        may be helpful to set a consistent name for specific channels, such as
        "in_transition", instead of using the autogenerated channel name, which
        can change between tunes. Any additional field names used will remain
        in the influx database, so be wary of programatically adding many of
        them.

        Args
        ------
        chan_info : list
            List of channel info corresponding to channels to monitor.
            Entries of this list can be:

              - ints: This will be interpreted as the readout-channel to
                monitor. The field-name will be set to "r<chan_no>".
              - tuples of length 2: Here the first element will be the readout
                chan to monitor, and the second value will be the field name to
                use for that channel.

            This list can be no more than 6 elements to limit how much detector
            data is saved to the HK format.

        sample_rate : float
            Target sample rate (Hz) to downsample detector data to. This must
            be less than 20 Hz to limit how much detector data is saved to hk.

        """
        monitored_chans = []
        for ch_info in params['chan_info']:
            if isinstance(ch_info, int):
                field_name = f"r{ch_info:0>4}"
                monitored_chans.append((ch_info, field_name))
            elif isinstance(ch_info, (tuple, list)):
                monitored_chans.append(ch_info)
            else:
                raise ValueError(
                    f"ch_info (type {type(ch_info)}) must be of type int or "
                    "tuple"
                )

        self.monitored_channels = monitored_chans
        self.monitored_chan_sample_rate = params['sample_rate']
        return True, "Set monitored channels"

    def _process_status(self, frame):
        """
        Processes a status frame. This will set or update the channel
        mask whenever the smurf metadata is updated.
        """
        if 'session_id' not in frame:
            return
        if self.mask_register in frame['status']:
            status = yaml.safe_load(frame['status'])
            self.mask = np.array(
                ast.literal_eval(status[self.mask_register])
            )

    def _process_monitored_chans(self, times, data):
        """
        Downsamples data for monitored channels and publishes tods to a grafana
        feed.
        """
        if not self.monitored_channels:
            return

        target_rate = self.monitored_chan_sample_rate
        if len(times) <= 1:
            ds_factor = 1
        else:
            input_rate = 1. / np.median(np.diff(times))
            ds_factor = max(int(input_rate // target_rate), 1)

        sl = slice(None, None, ds_factor)

        times_out = times[sl]
        for rc, field_name in self.monitored_channels:
            if rc >= len(data):
                self.log.warn(
                    f"Readout channel {rc} is larger than the number of"
                    f"streamed channels ({len(data)})! Data won't be published"
                    "to grafana."
                )
                continue
            _data = {
                'timestamps': times_out.tolist(),
                'block_name': field_name,
                'data': {
                    field_name: data[rc, sl].tolist()
                }
            }
            self.agent.publish_to_feed('detector_tods', _data)

    def _publish_wls(self, wls):
        """
        Publishes white-noise quantiles to ocs feed
        """
        quantiles = [15, 25, 50, 75, 85]
        labels = [f'white_noise_q{q}' for q in quantiles]
        data = {
            'timestamp': time.time(),
            'block_name': 'white_noise',
            'data': {
                k: np.quantile(wls, q / 100)
                for k, q in zip(labels, quantiles)
            }
        }
        self.agent.publish_to_feed('white_noise', data)

    def _process_data(self, frame, source_offset=0):
        """
        Processes a Scan frame. If lyrebird is enabled, this will return a seq
        of G3Frames that are formatted for lyrebird to ingest.
        """
        if 'session_id' not in frame:
            return []

        # Calculate downsample factor
        times_in, data_in = load_frame_data(frame)
        times_in = times_in - source_offset
        sample_rate = 1. / np.median(np.diff(times_in))
        nsamps = len(times_in)
        nchans = len(data_in)

        self._process_monitored_chans(times_in, data_in)

        if self.demod is None:
            self.demod = Demodulator(self.demod_freq, self.demod_bandwidth,
                                     fs=sample_rate)
        if self.wncalc is None:
            self.wncalc = WhiteNoiseCalculator(
                fs=sample_rate, navg=int(sample_rate)
            )

        demod = self.demod.apply(times_in, data_in)
        # white noise in units of pA/rt(Hz)
        wl = self.wncalc.apply(data_in * pA_per_rad)
        self._publish_wls(np.median(wl, axis=1))

        ds_factor = sample_rate // self.target_rate
        if np.isnan(ds_factor):  # There is only one element in the timestream
            ds_factor = 1
        ds_factor = max(int(ds_factor), 1)  # Prevents downsample factors < 1

        # Arrange output data structure
        sample_idxs = np.arange(self.ds_offset, nsamps, ds_factor, dtype=np.int32)
        self.ds_offset = ds_factor - (nsamps - self.ds_offset) % ds_factor
        num_frames = len(sample_idxs)

        times_out = times_in[sample_idxs]
        abs_chans = self.mask[np.arange(nchans)]

        nelems = len(self.fp.channels)
        raw_out = np.zeros((num_frames, nelems))
        demod_out = np.zeros((num_frames, nelems))
        wl_out = np.zeros((num_frames, nelems))

        for i, c in enumerate(abs_chans):
            if c >= len(self.fp.chan_mask):
                continue
            idx = self.fp.chan_mask[c]
            if idx >= 0:
                raw_out[:, idx] = data_in[i, sample_idxs]
                demod_out[:, idx] = demod[i, sample_idxs]
                wl_out[:, idx] = wl[i, sample_idxs]

        out = []
        for i in range(num_frames):
            fr = core.G3Frame(core.G3FrameType.Scan)
            fr['idx'] = 0
            fr['data'] = core.G3VectorDouble(raw_out[i])
            fr['timestamp'] = core.G3Time(times_out[i] * core.G3Units.s)
            out.append(fr)

            fr = core.G3Frame(core.G3FrameType.Scan)
            fr['idx'] = 1
            fr['data'] = core.G3VectorDouble(demod_out[i])
            fr['timestamp'] = core.G3Time(times_out[i] * core.G3Units.s)
            out.append(fr)

            fr = core.G3Frame(core.G3FrameType.Scan)
            fr['idx'] = 2
            fr['data'] = core.G3VectorDouble(wl_out[i])
            fr['timestamp'] = core.G3Time(times_out[i] * core.G3Units.s)
            out.append(fr)

        return out

    def read(self, session, params=None):
        """read(src='tcp://localhost:4532')

        **Process** - Process for reading in G3Frames from a source or list of
        sources. If this source is an address that begins with ``tcp://``, the
        agent will attempt to connect to a G3NetworkSender at the specified
        location. The ``src`` param can also be a filepath or list of filepaths
        pointing to G3Files to be streamed. If a list of filenames is passed,
        once the first file is finished streaming, subsequent files will be
        streamed.
        """

        self._running = True
        session.set_status('running')

        src_idx = 0
        if isinstance(params['src'], str):
            sources = [params['src']]
        else:
            sources = params['src']

        reader = None
        source = None
        source_offset = 0
        while self._running:

            if reader is None:
                try:
                    source = sources[src_idx]
                    source_is_file = not source.startswith('tcp://')
                    reader = core.G3Reader(source, timeout=5)
                except RuntimeError as e:
                    if source_is_file:
                        # Raise error if file cannot be found
                        raise e
                    else:
                        # If not a file, log error and try again
                        self.log.error("G3Reader could not connect! Retrying in 10 sec.")
                        time.sleep(10)
                        continue

            frames = reader.Process(None)
            if not frames:
                # If source is a file, start over with next file or break if
                # finished all sources. If socket, just reset reader and try to
                # reconnect
                if source_is_file:
                    src_idx += 1
                    if src_idx >= len(sources):
                        self.log.info("Finished reading all sources")
                        break
                reader = None
                continue

            frame = frames[0]

            # If this source is a file, this will shift the timestamps so that
            # data lines up with the current timestamp instead of using the
            # timestamps in the file
            if source_is_file and (not source_offset):
                source_offset = frame['time'].time / core.G3Units.s \
                    - time.time()
            elif not source_is_file:
                source_offset = 0

            if frame.type == core.G3FrameType.Wiring:
                self._process_status(frame)
                continue
            elif frame.type == core.G3FrameType.Scan:
                out = self._process_data(frame, source_offset=source_offset)
            else:
                continue

            for f in out:
                # This will block until there's a free spot in the queue.
                # This is useful if the src is a file and reader.Process does
                # not block
                self.out_queue.put(f)
        return True, "Stopped read process"

    def _stop_read(self, session, params=None):
        self._running = False
        return True, "Stopping read process"

    def stream_fake_data(self, session, params=None):
        """stream_fake_data()

        **Process** - Process for streaming fake data. This will queue up
        G3Frames full of fake data to be sent to lyrebird.
        """
        self._run_fake_stream = True
        ndets = len(self.fp.channels)
        chans = np.arange(ndets)
        frame_start = time.time()
        while self._run_fake_stream:
            time.sleep(2)
            frame_stop = time.time()
            ts = np.arange(frame_start, frame_stop, 1. / self.target_rate)
            frame_start = frame_stop
            nframes = len(ts)

            data_out = np.random.normal(0, 1, (nframes, ndets))
            data_out += np.sin(2 * np.pi * ts[:, None] + .2 * chans[None, :])

            for t, d in zip(ts, data_out):
                fr = core.G3Frame(core.G3FrameType.Scan)
                fr['idx'] = 0
                fr['data'] = core.G3VectorDouble(d)
                fr['timestamp'] = core.G3Time(t * core.G3Units.s)
                self.out_queue.put(fr)

                fr = core.G3Frame(core.G3FrameType.Scan)
                fr['idx'] = 1
                fr['data'] = core.G3VectorDouble(np.sin(d))
                fr['timestamp'] = core.G3Time(t * core.G3Units.s)
                self.out_queue.put(fr)

        return True, "Stopped fake stream process"

    def _stop_stream_fake_data(self, session, params=None):
        self._run_fake_stream = False
        return True, "Stopping fake stream process"

    @ocs_agent.param('dest', type=int)
    def send(self, session, params=None):
        """send(dest)

        **Process** - Process for sending outgoing G3Frames. This will query
        the out_queue for frames to be sent to lyrebird. This will try to
        regulate how fast it sends frames such that the delay between when the
        frames are sent, and the timestamp of the frames are fixed.

        """
        self._send_running = True

        first_frame_time = None
        stream_start_time = None

        sender = core.G3NetworkSender(
            hostname='*', port=params['dest'], max_queue_size=1000
        )

        sender.Process(self.fp.config_frame())
        session.set_status('running')
        while session.status in ['starting', 'running']:
            f = self.out_queue.get(block=True)
            t = f['timestamp'].time / core.G3Units.s
            now = time.time()
            if first_frame_time is None:
                first_frame_time = t
                stream_start_time = now

            this_frame_time = stream_start_time + (t - first_frame_time) + self.delay
            sleep_while_running(this_frame_time - now, session)
            sender.Process(f)

        return True, "Stopped send process"

    def _send_stop(self, session, params=None):
        self._send_running = False
        return True, "Stopping send process"


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--src', nargs='+', default='tcp://localhost:4532',
                        help="Address of incoming G3Frames.")
    pgroup.add_argument('--dest', type=int, default=8675,
                        help="Port to serve lyrebird frames")
    pgroup.add_argument('--stream-id', type=str, default='none',
                        help="Stream-id to use to distinguish magpie streams."
                             "This will be prepended to data-val names in lyrebird.")
    pgroup.add_argument('--target-rate', '-t', type=float, default=20,
                        help="Target sample rate for data being sent to Lyrebird. "
                             "Detector data will be downsampled to this rate.")
    pgroup.add_argument(
        '--delay', type=float, default=5,
        help="Delay (sec) between the timestamp of a G3Frame relative to the "
             "initial frame, and when the frame should be sent to lyrebird. "
             "This must be larger than the frame-aggregation time for smooth "
             "update times in lyrebird."
    )
    pgroup.add_argument('--layout', '-l', default='grid', choices=['grid', 'wafer'],
                        help="Focal plane layout style")
    pgroup.add_argument('--xdim', type=int, default=64,
                        help="Number of pixels in x-dimension for grid layout")
    pgroup.add_argument('--ydim', type=int, default=64,
                        help="Number of pixels in y-dimension for grid layout")
    pgroup.add_argument('--wafer-scale', '--ws', type=float, default=50.,
                        help="scale of wafer coordinates")
    pgroup.add_argument('--det-map', type=str, help="Path to det-map csv file")
    pgroup.add_argument('--fake-data', action='store_true',
                        help="If set, will stream fake data instead of listening to "
                             "a G3stream.")
    pgroup.add_argument('--offset', nargs=2, default=[0, 0], type=float,
                        help="Offset of detector coordinates with respect to "
                             "lyrebird coordinate system")
    pgroup.add_argument('--monitored-channels', nargs='+', type=int, default=[],
                        help="Readout channels to start monitoring on startup")
    pgroup.add_argument('--monitored-channel-rate', type=float, default=10,
                        help="Target sample rate for monitored channels")
    pgroup.add_argument('--demod-freq', type=float, default=8,
                        help="Demodulation frequency")
    pgroup.add_argument('--demod-bandwidth', type=float, default=0.5,
                        help="Demodulation bandwidth")
    return parser


def main(args=None):
    txaio.use_twisted()
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))
    parser = make_parser()
    args = site_config.parse_args(agent_class='MagpieAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    magpie = MagpieAgent(agent, args)

    if args.fake_data:
        read_startup = False
    else:
        read_startup = {'src': args.src}

    agent.register_process('read', magpie.read, magpie._stop_read,
                           startup=read_startup)
    agent.register_process('stream_fake_data', magpie.stream_fake_data,
                           magpie._stop_stream_fake_data, startup=args.fake_data)
    agent.register_process('send', magpie.send, magpie._send_stop,
                           startup={'dest': args.dest})
    agent.register_task('set_target_rate', magpie.set_target_rate)
    agent.register_task('set_delay', magpie.set_delay)
    agent.register_task(
        'set_monitored_channels', magpie.set_monitored_channels,
        startup={'chan_info': args.monitored_channels,
                 'sample_rate': args.monitored_channel_rate}
    )

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
# contributors: zatkins, bkoopman, sbhimani, zhuber

import math
import socket
import sys
import time

import numpy as np

# helper dicts
sensor_key = {
    '0': 'disabled',
    '1': 'diode',
    '2': 'platinum rtd',
    '3': 'ntc rtd',
    '4': 'thermocouple',
    '5': 'capacitance'
}
sensor_lock = {v: k for k, v in sensor_key.items()}

units_key = {
    '1': 'kelvin',
    '2': 'celsius',
    '3': 'sensor'
}
units_lock = {v: k for k, v in units_key.items()}

tempco_key = {
    '1': 'negative',
    '2': 'positive'
}
tempco_lock = {v: k for k, v in tempco_key.items()}

format_key = {
    '1': "mV/K (linear)",
    '2': "V/K (linear)",
    '3': "Ohm/K (linear)",
    '4': "log Ohm/K (linear)"
}
format_lock = {v: k for k, v in format_key.items()}

output_modes_key = {
    '0': 'off',
    '1': 'closed loop',
    '2': 'zone',
    '3': 'open loop',
    '4': 'monitor out',
    '5': 'Warm up'
}
output_modes_lock = {v.lower(): k for k, v in output_modes_key.items()}

channel_key = {
    '0': 'none',
    '1': 'A',
    '2': 'B',
    '3': 'C',
    '4': 'D',
    '5': 'D2',
    '6': 'D3',
    '7': 'D4',
    '8': 'D5'
}
channel_lock = {v: k for k, v in channel_key.items()}

heater_display_key = {
    '1': 'current',
    '2': 'power'
}
heater_display_lock = {v: k for k, v in heater_display_key.items()}

max_current_key = {
    '0': 'User',
    '1': .707,
    '2': 1.,
    '3': 1.141,
    '4': 2.
}
max_current_lock = {v: k for k, v in max_current_key.items()}

heater_range_key = {
    "0": "off",
    "1": "low",
    "2": "medium",
    "3": "high"
}
heater_range_lock = {v: k for k, v in heater_range_key.items()}

ramp_key = {
    '0': 'off',
    '1': 'on'
}
ramp_lock = {v: k for k, v in ramp_key.items()}

# main class - Lakeshore 336 driver


class LS336:
    """
    Implements a lakeshore 336 box to interface with client scripts.
    Only contains locally relevant information; namely, port parameters.
    The state of the device is not stored locally to avoid the potential
    for inconsistent information.
    Device status can always be accessed (accurately) through a msg.
    """
    # Constructor and instance variables

    def __init__(self, ip, timeout=10):

        # LS336 defaults
        self.com = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.com.connect((ip, 7777))
        self.com.settimeout(timeout)

        self.timeout = timeout
        self.id = self.get_id()
        print(self.id)  # print idenfitication information to see if working

        # Get Channels
        # Test whether device has extra scanner installed first
        self.extra_scanner = False
        temps = self.get_kelvin('0')

        if len(temps) == 8:
            inps = ['A', 'B', 'C', 'D', 'D2', 'D3', 'D4', 'D5']
            self.extra_scanner = True
            self.channels = {inp: Channel(self, inp) for inp in inps}
        elif len(temps) == 4:
            inps = ['A', 'B', 'C', 'D']
            self.channels = {inp: Channel(self, inp) for inp in inps}
        else:
            raise ValueError("Can't determine number of channels. "
                             "Please debug.")

        # Get Heaters
        htrs = ['1', '2']
        self.heaters = {out: Heater(self, out) for out in htrs}

    # Instance methods

    # copied from Lakeshore372 driver on 2020/12/09,
    # modified end-of-method sleep to 0.1s in all cases
    def msg(self, message):
        """Send message to the Lakeshore 336 over ethernet.

        If we're asking for something from the Lakeshore (indicated by a ? in
        the message string), then we will attempt to ask twice before giving
        up due to potential communication timeouts.

        Parameters
        ----------
        message : str
            Message string as described in the Lakeshore 336 manual.

        Returns
        -------
        str
            Response string from the Lakeshore, if any. Else, an empty string.

        """
        msg_str = f'{message}\r\n'.encode()

        if '?' in message:
            self.com.send(msg_str)
            # Try once, if we timeout, try again.
            # Usually gets around single event glitches.
            for attempt in range(2):
                try:
                    time.sleep(0.061)
                    resp = str(self.com.recv(4096), 'utf-8').strip()
                    break
                except socket.timeout:
                    print("Warning: Caught timeout waiting for response "
                          "to '%s', trying again before giving up" % message)
                    if attempt == 1:
                        raise RuntimeError('Query response to Lakeshore timed '
                                           'out after two attempts. '
                                           'Check connection.')
        else:
            self.com.send(msg_str)
            resp = ''

        # No comms for 100ms after sending message (manual says 50ms)
        time.sleep(0.1)
        return resp

    def get_id(self):
        """Get identification information of the Lakeshore module"""
        return self.msg('*IDN?')

    def get_kelvin(self, inp):
        """Return a temperature reading of the specified input
        ('A', 'B', 'C', or 'D') or '0' for all inputs. If the
        extra 3062 scanner is installed, possible options are
        ('A', 'B', 'C', 'D', 'D1', 'D2', 'D3', 'D4, and 'D5').
        Note that D and D1 refer to the same channel!

        Parameters
        ----------
        inp : str
            channel to query

        Returns
        -------
        array or float
            array of four (or eight) floats if input is '0',
            float otherwise of temperature reading

        Raises
        ------
        ValueError
            Invalid input channel arguments
        """
        if self.extra_scanner:
            if inp not in ['0', 'A', 'B', 'C', 'D', 'D1',
                           'D2', 'D3', 'D4', 'D5']:
                raise ValueError(f'invalid input in msg_kelvin: {inp}')
        else:
            if inp not in ['0', 'A', 'B', 'C', 'D']:
                raise ValueError(f'invalid input in msg_kelvin: {inp}')

        resp = self.msg(f'KRDG? {inp}')

        if inp == '0':
            # casts to array of floats
            return np.array(np.char.split(resp, sep=',').item()).astype(float)
        else:
            return float(resp)

    def get_sensor(self, inp):
        """Return a sensor reading of the specified input
        ('A', 'B', 'C', or 'D') or '0' for all inputs. If the
        extra 3062 scanner is installed, possible options are
        ('A', 'B', 'C', 'D', 'D1', 'D2', 'D3', 'D4, and 'D5').
        Note that D and D1 refer to the same channel!

        Parameters
        ----------
        inp : str
            channel to query

        Returns
        -------
        array or float
            array of four (or eight) floats if input is '0',
            float otherwise of sensor reading

        Raises
        ------
        ValueError
            Invalid input channel arguments
        """
        if self.extra_scanner:
            if inp not in ['0', 'A', 'B', 'C', 'D', 'D1',
                           'D2', 'D3', 'D4', 'D5']:
                raise ValueError(f'invalid input in msg_kelvin: {inp}')
        else:
            if inp not in ['0', 'A', 'B', 'C', 'D']:
                raise ValueError(f'invalid input in msg_kelvin: {inp}')

        resp = self.msg(f'SRDG? {inp}')

        if inp == '0':
            # casts to array of floats
            return np.array(np.char.split(resp, sep=',').item()).astype(float)
        else:
            return float(resp)

    # def get_heater_range(self, htr):
    #     return self.heaters[htr].get_heater_range()

    # def get_max_current(self, htr):
    #     return self.heaters[htr].get_max_current()

    # def get_heater_percent(self, htr):
    #     return self.heaters[htr].get_heater_percent()

    # def get_setpoint(self, htr):
    #     return self.heaters[htr].get_setpoint()


class Channel:
    """Channel class for LS336

    Parameters
    ----------
    ls : LS336 object
        The parent LS336 device
    inp : str
        The channel we are building ('A', 'B', 'C', or 'D')
        Could also be 'D1','D2','D3','D4', or 'D5' if the extra
        Lakeshore 3062 scanner is installed on the LS336.
        D and D1 refer to the same channel!
    """

    def __init__(self, ls, inp):
        assert inp in ['A', 'B', 'C', 'D', 'D1', 'D2', 'D3', 'D4', 'D5']

        self.ls = ls
        self.input = inp
        self.num = int(channel_lock[self.input])
        self.get_input_type()
        self.get_input_curve()
        self.get_input_name()
        self.get_T_limit()

    def get_input_type(self):
        """Return sensor metadata, <sensor type>, <autorange>, <range>,
        <compensation>, and <units>. For diodes, only sensor type and
        units are relevant.

        Returns
        -------
        list
            <sensor type>, <autorange>, <range>, <compensation>, and <units>
        """
        resp = self.ls.msg(f'INTYPE? {self.input}').split(',')

        self.sensor_type = sensor_key[resp[0]]
        self.autorange = int(resp[1])
        self.range = int(resp[2])
        self.compensation = int(resp[3])
        self.units = units_key[resp[4]]

        return resp

    def _set_input_type(self, params):
        """Assign sensor metadata, <sensor type>, <autorange>, <range>,
        <compensation>, and <units>. For diodes, only sensor type and
        units are relevant.

        Parameters
        ----------
        params : list
            <sensor type>, <autorange>, <range>, <compensation>, and <units>
        """
        assert len(params) == 5

        reply = [str(self.input)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg(f"INTYPE {param_str}")

    def get_sensor_type(self):
        """Get the sensor type of the channel in plain text"""
        self.get_input_type()
        return self.sensor_type

    def set_sensor_type(self, type):
        """Set the sensor type on the channel

        Parameters
        ----------
        type : str
            Sensor type must be in 'Disabled', 'Diode', 'Platinum RTD',
            'NTC RTD', 'Thermocouple', 'Capacitance'
        """
        assert type.lower() in sensor_lock

        resp = self.get_input_type()
        resp[1] = sensor_lock[type.lower()]
        self.sensor_type = type.lower()
        return self._set_input_type(resp)

    def get_units(self):
        """Get the channel preferred units as plain text"""
        self.get_input_type()
        return self.units

    def set_units(self, units):
        """Set the channel preferred units

        Parameters
        ----------
        unit : str
            Channel preferred units must be in 'Kelvin', 'Celsius, 'Sensor
        """
        assert units.lower() in units_lock

        resp = self.get_input_type()
        resp[4] = units_lock[units.lower()]
        self.units = units.lower()
        return self._set_input_type(resp)

    def get_input_name(self):
        """Get the name of the channel shown on front display

        Returns
        -------
        str
            The channel name for display purposes
        """
        resp = self.ls.msg(f'INNAME? {self.input}').strip()
        self.input_name = resp
        return self.input_name

    def set_input_name(self, name):
        """Set the name of the channel shown on front display

        Parameters
        ----------
        name : str
            The channel name for display purposes. Only send the first 15
            characters
        """
        name = name[:15]
        self.input_name = name
        resp = self.ls.msg(f'INNAME {self.input},{name}')
        return resp

    def get_input_curve(self):
        """Return the curve number of the curve assigned to this channel"""
        resp = self.ls.msg(f'INCRV? {self.input}').strip()
        self.input_curve = int(resp)
        return self.input_curve

    def set_input_curve(self, curve_num):
        """Set the curve number of the curve assigned to this channel"""
        assert curve_num in range(1, 60)

        self.input_curve = curve_num
        resp = self.ls.msg(f'INCRV {self.input},{self.input_curve}')
        return resp

    def get_T_limit(self):
        """Return the temperature limit above which control outputs assigned
        this channel shut off"""
        resp = self.ls.msg(f'TLIMIT? {self.input}')
        self.T_limit = float(resp)
        return self.T_limit

    def set_T_limit(self, limit):
        """Set the temperature limit above which control outputs assigned
        this channel shut off"""
        self.T_limit = limit
        resp = self.ls.msg(f'TLIMIT {self.input},{self.T_limit}')
        return resp

# Curve class copied from socs/Lakeshore372.py on 2020/08/21


class Curve:
    """Calibration Curve class for the LS336."""

    def __init__(self, ls, curve_num):
        self.ls = ls
        self.curve_num = curve_num
        self.get_header()

    def get_header(self):
        """Get curve header description.

        Returns
        -------
        list
            response from CRVHDR? in list

        """
        resp = self.ls.msg(f"CRVHDR? {self.curve_num}").split(',')
        print(resp)

        _name = resp[0].strip()
        _sn = resp[1].strip()
        _format = resp[2]
        _limit = float(resp[3])
        _coefficient = resp[4]

        self.name = _name
        self.serial_number = _sn

        self.format = format_key[_format]

        self.limit = _limit
        self.coefficient = tempco_key[_coefficient]

        return resp

    def _set_header(self, params):
        """Set the Curve Header with the CRVHDR command.

        Parameters should be <name>, <SN>, <format>, <limit value>,
        <coefficient>. We will determine <curve> from attributes. This
        allows us to use output from get_header directly, as it doesn't return
        the curve number.

        <name> is limited to 15 characters. Longer names take the first
               15 characters
        <sn> is limited to 10 characters. Longer sn's take the last 10 digits

        Parameters
        ----------
        params : list
            CRVHDR parameters

        Returns
        -------
        str
            response from ls.msg

        """
        assert len(params) == 5

        _curve_num = self.curve_num
        _name = params[0][:15]
        _sn = params[1][-10:]
        _format = params[2]
        assert _format.strip() in ['1', '2', '3', '4']
        _limit = params[3]
        _coeff = params[4]
        assert _coeff.strip() in ['1', '2']

        print(f'CRVHDR {_curve_num},{_name},{_sn},{_format},{_limit},{_coeff}')
        return self.ls.msg(f'CRVHDR {_curve_num},{_name},{_sn},'
                           f'{_format},{_limit},{_coeff}')

    def get_name(self):
        """Get the curve name with the CRVHDR? command.

        Returns
        -------
        str
            The curve name

        """
        self.get_header()
        return self.name

    def set_name(self, name):
        """Set the curve name with the CRVHDR command.

        Parameters
        ----------
        name : str
            The curve name, limit of 15 characters, longer names get
            truncated

        Returns
        -------
        str
            the response from the CRVHDR command

        """
        resp = self.get_header()
        resp[0] = name.upper()
        self.name = resp[0]
        return self._set_header(resp)

    def get_serial_number(self):
        """Get the curve serial number with the CRVHDR? command."

        Returns
        -------
        str
            The curve serial number

        """
        self.get_header()
        return self.serial_number

    def set_serial_number(self, serial_number):
        """Set the curve serial number with the CRVHDR command.

        Parameters
        ----------
        serial_number : str
            The curve serial number, limit of 10 characters, longer serials get
            truncated

        Returns
        -------
        str
            the response from the CRVHDR command

        """
        resp = self.get_header()
        resp[1] = serial_number
        self.serial_number = resp[1]
        return self._set_header(resp)

    def get_format(self):
        """Get the curve data format with the CRVHDR? command."

        Returns
        -------
        str
            The curve data format

        """
        self.get_header()
        return self.format

    def set_format(self, _format):
        """Set the curve format with the CRVHDR command.

        Parameters
        ----------
        _format : str
            The curve format, valid formats are: "mV/K (linear)", "V/K
            (linear)", "Ohm/K (linear)", and "log Ohm/K (linear)"

        Returns
        -------
        str
            the response from the CRVHDR command

        """
        resp = self.get_header()

        assert _format in format_lock.keys(), "Please select a valid format"

        resp[2] = format_lock[_format]
        self.format = _format
        return self._set_header(resp)

    def get_limit(self):
        """Get the curve temperature limit with the CRVHDR? command.

        Returns
        -------
        str
            The curve temperature limit

        """
        self.get_header()
        return float(self.limit)

    def set_limit(self, limit):
        """Set the curve temperature limit with the CRVHDR command.

        Parameters
        ----------
        limit : float
            The curve temperature limit

        Returns
        -------
        str
            the response from the CRVHDR command

        """
        resp = self.get_header()
        resp[3] = str(limit)
        self.limit = limit
        return self._set_header(resp)

    def get_coefficient(self):
        """Get the curve temperature coefficient with the CRVHDR? command.

        Returns
        -------
        str
            The curve temperature coefficient

        """
        self.get_header()
        return self.coefficient

    def set_coefficient(self, coefficient):
        """Set the curve temperature coefficient with the CRVHDR command.

        Parameters
        ----------
        coefficient : str
            The curve temperature coefficient, either 'positive' or 'negative'

        Returns
        -------
        str
            the response from the CRVHDR command

        """
        assert coefficient in ['positive', 'negative']

        resp = self.get_header()
        resp[4] = tempco_lock[coefficient]
        self.tempco = coefficient
        return self._set_header(resp)

    def get_data_point(self, index):
        """Get a single data point from a curve, given the index, using the
        CRVPT? command.

        Parameters
        ----------
        index : int
            index of breakpoint to msg

        Returns
        -------
        tuple
            (units, tempertaure, curvature) values for the given breakpoint

            The format for the return value, a 2-tuple of floats, is chosen to work
            with how the get_curve() method later stores the entire curve in a
            numpy structured array.

        """
        resp = self.ls.msg(f"CRVPT? {self.curve_num},{index}").split(',')
        _units = float(resp[0])
        _temp = float(resp[1])
        return (_units, _temp)

    def _set_data_point(self, index, units, kelvin):
        """Set a single data point with the CRVPT command.

        Parameters
        ----------
        index : int
            data point index
        units : float
            value of the sensor units to 6 digits
        kelvin : float
            value of the corresponding temp in Kelvin to 6 digits

        Returns
        -------
        str
            response from the CRVPT command

        """
        resp = self.ls.msg(
            f"CRVPT {self.curve_num}, {index}, {units}, {kelvin}")

        return resp

    # Public API Elements
    def get_curve(self, _file=None):
        """Get a calibration curve from the LS336.
        If _file is not None, save to file location.
        """
        breakpoints = []
        for i in range(1, 201):
            x = self.get_data_point(i)
            if x[0] == 0:
                break
            breakpoints.append(x)

        struct_array = np.array(breakpoints, dtype=[('units', 'f8'),
                                                    ('temperature', 'f8')])

        self.breakpoints = struct_array

        if _file is not None:
            with open(_file, 'w') as f:
                f.write('Curve Name:\t' + self.name + '\r\n')
                f.write('Serial Number:\t' + self.serial_number + '\r\n')
                f.write('Data Format:\t'
                        + format_lock[self.format] + f'\t({self.format})\r\n')
                f.write('SetPoint Limit:\t%s\t(Kelvin)\r\n' % '%0.4f' %
                        np.max(self.breakpoints['temperature']))
                f.write('Temperature coefficient:\t'
                        + tempco_lock[self.coefficient]
                        + f' ({self.coefficient})\r\n')
                f.write('Number of Breakpoints:\t%s\r\n' %
                        len(self.breakpoints))
                f.write('\r\n')
                f.write('No.\tUnits\tTemperature (K)\r\n')
                f.write('\r\n')
                for idx, point in enumerate(self.breakpoints):
                    f.write('%s\t%s %s\r\n' % (idx + 1, '%0.5f' %
                                               point['units'],
                                               '%0.4f' % point['temperature']))

        return self.breakpoints

    def set_curve(self, _file):
        """Set a calibration curve, loading it from the file.

        Parameters
        ----------
        _file : str
            the file to load the calibration curve from

        Returns
        -------
        list
            return the new curve header, refreshing the attributes

        """
        with open(_file) as f:
            content = f.readlines()

        header = []
        for i in range(0, 6):
            if i < 2 or i > 4:
                header.append(content[i].strip().split(":", 1)[1].strip())
            else:
                header.append(content[i].strip().split(":", 1)[
                              1].strip().split("(", 1)[0].strip())

        # Skip to the V and T values in the file and strip them of tabs,
        # newlines, etc
        values = []
        for i in range(9, len(content)):
            values.append(content[i].strip().split())

        self.delete_curve()
        # remove old curve first, so old breakpoints don't remain
        time.sleep(1)  # necessary to make work

        self._set_header(header[:-1])  # ignore num of breakpoints

        for point in values:
            print("uploading %s" % point)
            self._set_data_point(point[0], point[1], point[2])

        # refresh curve attributes
        self.get_header()
        self._check_curve(_file)

    def _check_curve(self, _file):
        """After setting a data point for calibration curve,
        use CRVPT? command from get_data_point() to check
        that all points of calibration curve  were uploaded.
        If not, re-upload points.

        Parameters
        ----------
        _file : str
            calibration curve file

        """
        with open(_file) as f:
            content = f.readlines()

        # skipping header info
        values = []
        for i in range(9, len(content)):
            # data points that should have been uploaded
            values.append(content[i].strip().split())

        for j in range(1, len(values) + 1):
            try:
                resp = self.get_data_point(j)  # response from the 336
                point = values[j - 1]
                units = float(resp[0])
                temperature = float(resp[1])
                assert units == float(
                    point[1]), "Point number %s not uploaded" % point[0]
                assert temperature == float(
                    point[2]), "Point number %s not uploaded" % point[0]
                print("Successfully uploaded %s, %s" % (units, temperature))
            # if AssertionError, tell 336 to re-upload points
            except AssertionError:
                if units != float(point[1]):
                    self.set_curve(_file)

    def delete_curve(self):
        """Delete the curve using the CRVDEL command.

        Returns
        -------
        str
            the response from the CRVDEL command

        """
        resp = self.ls.msg(f"CRVDEL {self.curve_num}")
        # self.get_header()
        return resp

    def soft_cal(self, std, points, delay=1):
        """Executes SCAL command using the data in points_str.
        Note this overwrites the current breakpoints!

        Parameters
        ----------
        std : int
            Standard curve number to base the SoftCal on
        points : list
            List of T1,U1,T2,U2,T3,U3 values
        """
        assert len(points) == 6
        points_str = ','.join(points)

        resp = self.ls.msg(
            f'SCAL {std},{self.curve_num},{self.serial_number},{points_str}',
            delay=delay)
        self.get_header()
        return resp

    def __str__(self):
        string = "-" * 50 + "\n"
        string += "Curve %d: %s\n" % (self.curve_num, self.name)
        string += "-" * 50 + "\n"
        string += "  %-30s\t%r\n" % ("Serial Number:", self.serial_number)
        string += "  %-30s\t%s (%s)\n" % ("Format :",
                                          format_lock[self.format],
                                          self.format)
        string += "  %-30s\t%s\n" % ("Temperature Limit:", self.limit)
        string += "  %-30s\t%s\n" % ("Temperature Coefficient:",
                                     self.coefficient)

        return string

# Heater class copied from socs/Lakeshore372.py on 2020/08/25
# code modified for LS336 thereafter


class Heater:
    """Heater class for LS336 control

    Parameters
    ----------
    ls : Lakeshore336.LS336
        the lakeshore object we're controlling
    output : int
        the heater output we want to control, 1 = 100W, 2 = 50W

    """

    def __init__(self, ls, output):
        assert int(output) in [1, 2]

        self.ls = ls
        self.output = output
        self.output_name = f'Heater {output}'
        self.resistance = None

        self.get_output_mode()
        self.get_heater_setup()
        self.get_heater_range()
        self.get_setpoint()

    def get_output_mode(self):
        """msg the heater mode using the OUTMODE? command.

        Returns
        -------
        tuple
            3-tuple with output mode, input, and whether powerup is enabled

        """
        resp = self.ls.msg(f"OUTMODE? {self.output}").split(",")

        # TODO: make these human readable
        self.mode = output_modes_key[resp[0]]
        self.input = channel_key[resp[1]]
        self.powerup = resp[2]
        return resp

    # OUTMODE
    def _set_output_mode(self, params):
        """Set the output mode of the heater with the OUTMODE command.
        Parameters should be <mode>, <input>, and <powerup enable>.
        This allows us to use output from get_output_mode directly, as
        it doesn't return <output>.

        Parameters
        ----------
        params : list
            OUTMODE parameters

        Returns
        -------
        str
            response from ls.msg

        """
        assert len(params) == 3

        reply = [str(self.output)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg(f"OUTMODE {param_str}")

    def get_mode(self):
        """Set output mode with OUTMODE? commnd.

        Returns
        -------
        str
            The output mode

        """
        self.get_output_mode()
        return self.mode

    def set_mode(self, mode):
        """Set output mode with OUTMODE commnd.

        Parameters
        ----------
        mode : str
            control mode for heater

        Returns
        -------
        str
            the response from the OUTMODE command

        """
        assert mode.lower() in output_modes_lock.keys(
        ), f"{mode} not a valid mode"

        resp = self.get_output_mode()
        resp[0] = output_modes_lock[mode.lower()]
        self.mode = mode
        return self._set_output_mode(resp)

    def get_input_channel(self):
        """Get the control channel with the OUTMODE? command.

        Returns
        -------
        str
            The control channel

        """
        self.get_output_mode()
        return self.input

    def set_input_channel(self, inp):
        """Set the control channel with the OUTMODE command.

        Parameters
        ----------
        inp : str or int
            specifies which input or channel to control from

        """
        assert inp in channel_lock, f"{inp} not a valid input/channel"

        resp = self.get_output_mode()
        resp[1] = channel_lock[str(inp)]
        self.input = str(inp)
        return self._set_output_mode(resp)

    def get_powerup(self):
        """Get the powerup state with the OUTMODE? command.

        Returns
        -------
        str
            The powerup state

        """
        self.get_output_mode()
        return self.powerup

    def set_powerup(self, powerup):
        """
        Parameters
        ----------
        powerup : bool
            specifies whether the output remains on or shuts off after power
            cycle. True for on after powerup

        """
        assert powerup in [
            True, False], f"{powerup} not valid powerup parameter"

        resp = self.get_output_mode()
        set_powerup = str(int(powerup))
        resp[2] = set_powerup
        self.powerup = set_powerup
        return self._set_output_mode(resp)

    def get_heater_setup(self):
        """Gets Heater setup params with the HTRSET? command.

        Returns
        -------
        list
            List of values that have been returned from the Lakeshore.

        """
        resp = self.ls.msg("HTRSET? {}".format(self.output)).split(',')

        self.resistance_setting = int(resp[0])
        self.max_current = max_current_key[resp[1]]
        self.max_user_current = float(resp[2].strip('E+'))
        self.display = heater_display_key[resp[3]]

        return resp

    def _set_heater_setup(self, params):
        """
        Sets the heater setup using the HTRSET command.

        Parameters
        ----------
        params : list
            Params must be a list with the parameters:
                <heater resistance mode>:    Heater mode in Ohms; 1=25 Ohms,
                                             2=50 Ohms
                <max current>: Specifies max heater output for warm-up heater.
                               0=User spec, 1=0.707 A, 2=1 A, 3=1.141 A, 4=2 A.
                <max user current>: Max heater output if max_current is set to user
                <current/power>:    Specifies if heater display is current or
                                    power. 1=current, 2=power.

        """
        assert len(params) == 4

        reply = [str(self.output)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg("HTRSET {}".format(param_str))

    def get_heater_resistance_setting(self):
        """Get the "setting" of the heater resistance, which can only be
        25 or 50 Ohms
        """
        self.get_heater_setup()
        if self.resistance is None:
            if self.resistance_setting == 1:
                self.resistance = 25
            elif self.resistance_setting == 2:
                self.resistance = 50
        return self.resistance_setting

    def set_heater_resistance(self, res):
        """Sets the correct heater setting depending on the actual
        load resistance

        Parameters
        ----------
        res : int or float
            Actual resistance of load being powered
        """
        if res < 50:
            setting = 1  # 25 Ohm
        elif res >= 50:
            setting = 2  # 50 Ohm
        self.resistance = res

        resp = self.get_heater_setup()
        resp[0] = str(setting)
        self.resistance_setting = setting
        return self._set_heater_setup(resp)

    def get_max_current(self):
        """Get the limiting current of the heater. Either set by "max current"
        if user max current not on, or user max current if user max current on.

        Returns
        -------
        float
            Limiting heater current
        """
        self.get_heater_setup()
        if self.max_current == 'User':
            return self.max_user_current
        else:
            return self.max_current

    def set_max_current(self, current):
        assert current <= 2, f'Current {current} is too high (>2 A)'
        # round down to 3 decimal places
        current = math.floor(1000 * current) / 1000

        resp = self.get_heater_setup()
        if current in max_current_lock:
            resp[1] = max_current_lock[current]
            self.max_current = current
        else:
            resp[1] = '0'
            self.max_current = 'User'
            resp[2] = str(current)
            self.max_user_current = current
        return self._set_heater_setup(resp)

    def get_heater_display(self):
        """Get whether heater displays in % of full current or power

        Returns
        -------
        str
            Display unit (% of current or % of power)
        """
        self.get_heater_setup()
        return self.display

    def set_heater_display(self, display):
        """Change the display of the heater

        Parameters
        ----------
        display : str
            Display mode for heater. Can either be 'current' or 'power'.

        """
        assert display.lower() in heater_display_lock.keys(
        ), f"{display} is not a valid display"

        resp = self.get_heater_setup()
        resp[3] = heater_display_lock[display.lower()]
        self.display = display
        return self._set_heater_setup(resp)

    def get_manual_out(self):
        """Return the % of full current or power depending on heater display,
        if set by MOUT

        Returns
        -------
        float
            the % of full current or power depending on heater display
        """
        resp = self.ls.msg("MOUT? {}".format(self.output))
        return float(resp)

    def set_manual_out(self, percent):
        """Set the % of full current or power depending on heater display,
        with MOUT

        Parameters
        ----------
        percent : int or float
            the % of full current or power depending on heater display
        """
        assert 100 * \
            percent == int(
                100 * percent), ("Percent value cannot have more than 2 "
                                 "decimal places")

        resp = self.ls.msg(f'MOUT {self.output},{percent}')
        return resp

    def get_heater_range(self):
        """Get heater range with RANGE? command.

        Returns
        -------
        float
            heater range by decade in total available power/current

        """
        resp = self.ls.msg(f"RANGE? {self.output}")
        self.range = heater_range_key[resp]
        return self.range

    def set_heater_range(self, rng):
        """Set heater range with RANGE command.

        Parameters
        ----------
        rng : str
            heater range, either 'Off','Low','Medium', or 'High'

        """
        _range = rng.lower()
        assert _range in heater_range_lock.keys(), 'Not a valid heater Range'

        self.range = heater_range_lock[_range]
        resp = self.ls.msg(f"RANGE {self.output},{heater_range_lock[_range]}")
        return resp

    def get_setpoint(self):
        """Return the setpoint in control loop sensor units"""
        resp = self.ls.msg(f"SETP? {self.output}")
        self.setpoint = float(resp)
        return self.setpoint

    def set_setpoint(self, value):
        """Set the setpoint of the control loop in sensor units

        Parameters
        ----------
        value : int or float
            The setpoint. Units depend on the preferred sensor units.
        """
        self.setpoint = float(value)
        resp = self.ls.msg(f"SETP {self.output},{value}")
        return resp

    def get_pid(self):
        """Get PID parameters with PID? command.

        Returns
        -------
        tuple
            (p, i, d)

        """
        resp = self.ls.msg(f"PID? {self.output}").split(',')
        self.p = float(resp[0])
        self.i = float(resp[1])
        self.d = float(resp[2])
        return self.p, self.i, self.d

    def set_pid(self, p, i, d):
        """Set PID parameters for closed loop control.

        Parameters
        ----------
        p : float
            proportional term in PID loop
        i : float
            integral term in PID loop
        d : float
            derivative term in PID loop

        Returns
        -------
        str
            response from PID command

        """
        assert float(p) <= 1000 and float(p) >= 0.1
        assert float(i) <= 1000 and float(i) >= 0.1
        assert float(d) <= 200 and float(d) >= 0

        self.p = p
        self.i = i
        self.d = d

        resp = self.ls.msg(f"PID {self.output},{p},{i},{d}")
        return resp

    def get_ramp(self):
        """Return list of params <on/off>, <rate value> for RAMP? msg

        Returns
        -------
        list
            <on/off>, <rate value> for RAMP? msg
        """
        resp = self.ls.msg(f'RAMP? {self.output}').split(',')
        self.ramp_enabled = ramp_key[resp[0]]
        self.ramp_rate = float(resp[1])
        return resp

    def _set_ramp(self, params):
        """Set the RAMP params <on/off>, <rate value>

        Parameters
        ----------
        params : list
            <on/off>, <rate value>
        """
        assert len(params) == 2

        reply = [str(self.output)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        resp = self.ls.msg(f'RAMP {param_str}')
        return resp

    def get_ramp_on_off(self):
        """Get string indicating ramp 'on' or 'off' """
        self.get_ramp()
        return self.ramp_enabled

    def set_ramp_on_off(self, on_off):
        """Turn ramp on or off

        Parameters
        ----------
        on_off : str
            Either 'on' to enable ramp or 'off' to disable ramp
        """
        assert on_off.lower() in ramp_lock

        resp = self.get_ramp()
        resp[0] = ramp_lock[on_off.lower()]
        self.ramp_enabled = on_off.lower()
        return self._set_ramp(resp)

    def get_ramp_rate(self):
        """Returns ramp rate in K/min"""
        self.get_ramp()
        return self.ramp_rate

    def set_ramp_rate(self, rate):
        """Set ramp rate of changes in setpoint, in K/min

        Parameters
        ----------
        rate : int or float
            Absolute value of setpoint rate of change, from 0.1 to 100 K/min
        """
        assert rate >= 0.1 and rate <= 100

        resp = self.get_ramp()
        resp[1] = str(rate)
        self.ramp_rate = rate
        return self._set_ramp(resp)

    def get_ramp_status(self):
        """Return a string indicating whether or not the setpoint
        is currently ramping"""
        resp = self.ls.msg(f'RAMPST? {self.output}')
        ramp_stat_dict = {'0': 'Not ramping', '1': 'Ramping'}
        self.ramp_status = ramp_stat_dict[resp]
        return self.ramp_status

    def get_heater_percent(self):
        """Return the current heater output level in %"""
        resp = self.ls.msg(f'HTR? {self.output}')
        self.percent = float(resp)
        return self.percent

# Do stuff


if __name__ == '__main__':

    # Initialize device from CLI

    port = sys.argv[1]
    ls = LS336(port)

    # Ask basic questions
    print('Lakeshore Initialized, SN:', ls.msg('*IDN?'))



"""Michael Randall
mrandall@ucsd.edu"""

import argparse
import socket
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.agents.tektronix3021c.drivers import TektronixInterface


class TektronixAWGAgent:
    """Tektronix3021c Function Generator Agent.

    Args:
        ip_address (string): The IP address of the gpib to ethernet
            controller connected to the function generator.
        gpib_slot (int): The gpib address currently set
            on the function generator.

    """

    def __init__(self, agent, ip_address, gpib_slot):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.job = None

        self.ip_address = ip_address
        self.gpib_slot = gpib_slot
        self.monitor = False

        self.awg = None
        # Registers data feeds
        agg_params = {
            'frame_length': 60,
        }
        self.agent.register_feed('awg',
                                 record=True,
                                 agg_params=agg_params)

    @ocs_agent.param('_')
    def init(self, session, params=None):
        """init()

        **Task** - Initialize connection to Tektronix AWG.

        """

        with self.lock.acquire_timeout(0) as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            try:
                self.awg = TektronixInterface(self.ip_address, self.gpib_slot)
                self.idn = self.awg.identify()

            except socket.timeout as e:
                self.log.error("""Tektronix AWG
                               timed out during connect -> {}""".format(e))
                return False, "Timeout"

            self.log.info("Connected to AWG: {}".format(self.idn))

        return True, 'Initialized AWG.'

    @ocs_agent.param('frequency', type=float, check=lambda x: 0 <= x <= 25e6)
    def set_frequency(self, session, params=None):
        """set_frequency(frequency)

        **Task** - Set frequency of the function generator.

        Parameters:
            frequency (float): Frequency to set in Hz. Must be between 0 and
                25,000,000.
        """

        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                freq = params['frequency']

                self.awg.set_freq(freq)

                data = {'timestamp': time.time(),
                        'block_name': "AWG_frequency_cmd",
                        'data': {'AWG_frequency_cmd': freq}
                        }
                self.agent.publish_to_feed('awg', data)
            else:
                return False, "Could not acquire lock"

        return True, 'Set frequency {} Hz'.format(freq)

    @ocs_agent.param('amplitude', type=float, check=lambda x: 0 <= x <= 10)
    def set_amplitude(self, session, params=None):
        """set_amplitude(amplitude)

        **Task** - Set peak to peak voltage of the function generator.

        Parameters:
            amplitude (float): Peak to Peak voltage to set. Must be between 0
                and 10.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                amp = params['amplitude']

                self.awg.set_amp(amp)

                data = {'timestamp': time.time(),
                        'block_name': "AWG_amplitude_cmd",
                        'data': {'AWG_amplitude_cmd': amp}
                        }
                self.agent.publish_to_feed('awg', data)
            else:
                return False, "Could not acquire lock"

        return True, 'Set amplitude to {} Vpp'.format(params)

    @ocs_agent.param('state', type=bool)
    def set_output(self, session, params=None):
        """set_output(state)

        **Task** - Turn function generator output on or off.

        Parameters:
            state (bool): True for on, False for off.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                state = params.get("state")

                self.awg.set_output(state)

                data = {'timestamp': time.time(),
                        'block_name': "AWG_output_cmd",
                        'data': {'AWG_output_cmd': int(state)}
                        }
                self.agent.publish_to_feed('awg', data)

            else:
                return False, "Could not acquire lock"

        return True, 'Set Output to {}.'.format(params)


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address', type=str,
                        help="IP address of Tektronix device.")
    pgroup.add_argument('--gpib-slot', type=int,
                        help="GPIB slot of Tektronix device.")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class="Tektronix AWG",
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    p = TektronixAWGAgent(agent, args.ip_address, args.gpib_slot)

    agent.register_task('init', p.init, startup=True)
    agent.register_task('set_frequency', p.set_frequency)
    agent.register_task('set_amplitude', p.set_amplitude)
    agent.register_task('set_output', p.set_output)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import os
import time

import numpy as np
import txaio
import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

ON_RTD = os.environ.get("READTHEDOCS") == "True"
if not ON_RTD:
    import casperfpga
    from holog_daq import fpga_daq3, synth3


class FPGAAgent:
    """
    Agent for programming FPGA and data acquisition for holography.

    Args:
        config_file (str): ocs-site-configs/uchicago/field/holog_config.yaml
    """

    def __init__(self, agent, config_file):

        self.fpga = None
        self.initialized = False
        # self.take_data = False

        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        agg_params = {"frame_length": 10 * 60}  # [sec]
        self.agent.register_feed(
            "fpga", record=True, agg_params=agg_params, buffer_time=0
        )

        # Load dictionary of specific mirror paramters, since some parameters
        # like limits and translate vary over different FTSes This is loaded
        # from a yaml file, which is assumed to be in the $OCS_CONFIG_DIR
        # directory.
        if config_file is None:
            raise Exception("No config file specified for the FTS mirror config")
        else:
            config_file_path = os.path.join(os.environ["OCS_CONFIG_DIR"], config_file)
            with open(config_file_path) as stream:
                self.holog_configs = yaml.safe_load(stream)
                if self.holog_configs is None:
                    raise Exception("No mirror configs in config file.")
                self.log.info(f"Loaded mirror configs from file {config_file_path}")
                self.baseline = self.holog_configs.pop("baseline", None)
                self.roach = self.holog_configs.pop("roach", None)
                self.path_to_roach_init = self.holog_configs.pop("path_to_roach_init", None)
                self.python2_env = self.holog_configs.pop("python2_env", None)
                # The other mirror configs (speed, timeout) are optional and
                # have defaults so we leave them as the dictionary.
                if self.roach is None or self.baseline is None:
                    raise Exception("IP address and channels must be included "
                                    "in the holography configuration keys.")

        self.roach, self.opts, self.baseline = fpga_daq3.roach2_init()

        err = os.system(self.python2_env + self.path_to_roach_init)
        assert err == 0

        print("Connecting to server %s ... " % (self.roach))

        self.fpga = casperfpga.CasperFpga(self.roach)
        time.sleep(1)

        if self.fpga.is_connected():
            print("ok\n")

    def take_data(self, session, params=None):
        """take_data()

        **Task** - A task to take data from the FPGA.

        Examples:
            Example for calling in a client::

                from ocs.ocs_client import OCSClient
                agent_fpga = OCSClient("fpga") # create agent
                agent_fpga.take_data() # take data

        Notes:
            An example of the session data::

                >>> response.session['data']
                {"timestamp": 1601924482.722671,
                 "block_name": "fpga",
                 "data": {"amp_AA": 293.644,
                          "amp_BB": 33.752,
                          "amp_AB": 33.752,
                          "arr_P": 33.752}
                }
        """
        with self.lock.acquire_timeout(timeout=3, job="take_data") as acquired:
            if not acquired:
                self.log.warn(
                    f"Could not set position because lock held by {self.lock.job}"
                )
                return False, "Could not acquire lock"

            # Grab synthesizer settings here (which FPGA bins to integrate over)
            self.synth_settings = synth3.SynthOpt()

            self.synth_settings.IGNORE_PEAKS_BELOW = int(655)
            self.synth_settings.IGNORE_PEAKS_ABOVE = int(660)
            # Take data here
            arr_aa, arr_bb, arr_ab, arr_phase, arr_index = fpga_daq3.TakeAvgData(
                self.baseline, self.fpga, self.synth_settings
            )

            # Data dictionary is what we will send to the data feed:
            data = {"timestamp": time.time(), "block_name": "fpga", "data": {}}

            arr_AA = np.array(fpga_daq3.running_mean(arr_aa.tolist(), 1))
            arr_BB = np.array(fpga_daq3.running_mean(arr_bb.tolist(), 1))
            arr_AB = np.array(fpga_daq3.running_mean(arr_ab.tolist(), 1))
            arr_P = np.array(fpga_daq3.running_mean(arr_phase.tolist(), 1))

            n_channels = np.size(arr_AA)

            amp_AA = arr_AA[int(n_channels / 2)]
            amp_BB = arr_BB[int(n_channels / 2)]
            amp_AB = np.power(arr_AB[int(n_channels / 2)], 1)
            amp_P = np.remainder(arr_P[int(n_channels / 2)], 360.0)

            data["data"]["amp_AA"] = amp_AA
            data["data"]["amp_BB"] = amp_BB
            data["data"]["amp_AB"] = amp_AB
            data["data"]["amp_P"] = amp_P

            self.agent.publish_to_feed("fpga", data)
            session.data.update(data["data"])

        return True, "Data acquired."


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--config_file")

    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class="FPGAAgent",
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    fpga_agent = FPGAAgent(agent, args.config_file)

    # agent.register_task("init_FPGA", fpga_agent.init_FPGA)
    agent.register_task("take_data", fpga_agent.take_data)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()
